# Copyright 2016-2018 The NOMAD Developers Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Maintainers:
# 2017-2018: Lorenzo Pardini <loren.pard@gmail.com>
# 2019-2020: Cuauhtemoc Salazar

from builtins import object
import numpy as np
from nomadcore.simple_parser import AncillaryParser, CachingLevel
from nomadcore.simple_parser import SimpleMatcher as SM, mainFunction
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.caching_backend import CachingLevel
from nomadcore.unit_conversion import unit_conversion
from nomadcore.unit_conversion.unit_conversion import convert_unit_function
import os, sys, json
import excitingparser.exciting_parser_dos as exciting_parser_dos
import excitingparser.exciting_parser_bandstructure as exciting_parser_bandstructure
import excitingparser.exciting_parser_gw as exciting_parser_gw
import excitingparser.exciting_parser_GS_input as exciting_parser_GS_input
import excitingparser.exciting_parser_XS_input as exciting_parser_XS_input
import excitingparser.exciting_parser_xs as exciting_parser_xs
import excitingparser.exciting_parser_eps as exciting_parser_eps
from ase import Atoms
import logging

logger = logging.getLogger("nomad.ExcitingParser")

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

class ExcitingParserContext(object):

  def __init__(self):
    self.parser = None
    # self.mainFileUri = sys.argv[1]    #exciting !!!!!!LOCAL HOME!!!!!!!!             OKOKOKOK
    self.mainFileUri = sys.argv[2]  #exciting !!! FOR NOMAD URI nmd:// or sbt -> zip file!!!!!!!!   OKOKKOOK
    self.mainFilePath = None
    self.mainFile = None
    self.volumeCubOpt = False
    self.volumeOpt = False
    self.volumeCubOptIndex = 0

  def initialize_values(self):
    self.metaInfoEnv = self.parser.parserBuilder.metaInfoEnv
    self.atom_pos = []
    self.atom_labels = []

  def startedParsing(self, path, parser):
    self.parser=parser
    self.initialize_values()
    # self.atom_pos = []
    # self.atom_labels = []
    self.XSSetGIndex = None
    self.secMethodIndex = None
    self.secSystemIndex = None
    self.secSingleConfIndex = None
    self.spinTreat = None
    self.sim_cell = []
    self.cell_format = ''
    self.secRunIndex = None
    self.unit_cell_vol = 0
    self.xcName = None
    self.gmaxvr = 0
    self.rgkmax = 0
    self.energy_thresh = []
    self.samplingMethod = None
    self.secSamplingMethodIndex = None
    self.geometryForceThreshold = 0
    self.frameSequence = []
    self.samplingGIndex = 0
    self.dummy = 0
    self.i=0
    self.clathrates = False
    self.bsetype = None
    self.screentype = None
    self.tensorComp = []
    self.excitonEnergies = []
    self.xsTetra = False
    self.xsAC = False
    self.xsNAR = False
    self.xstype = None
    self.tddftKernel = None
    # self.xsType = None
    # self.volumeOptIndex = 0

  def onOpen_section_run(self, backend, gIndex, section):
    #  self.i=self.i+1
    #  print("self.i=",self.i)
    curDir = os.getcwd()
    mainFile = self.parser.fIn.fIn.name           #####exciting sbt -> zip filei r from NOMAD URI nmd:// ####     YES ?????????? sure???? check first    OKOKOKKO
    self.mainFilePath = os.path.dirname(mainFile)           #####exciting sbt -> zip file####     YES ?????????? sure???? check first    OKOKOKKO
    #  self.mainFilePath = self.mainFileUri[0:-9]     #####exciting LOCAL HOME #######   YES                      OKOKOKOK
    #  print("self.mainFilePath===",self.mainFilePath)
    os.chdir(self.mainFilePath)
    os.chdir('../')
    #    self.i=self.i+1
    #    print("self.i=",self.i)
    #    self.volumeOptIndex = 0
    #    for root, dirs, files in os.walk('.'):
    #      if root == '.':
    #        for directory in dirs:
    #          if directory[-4:-1]==
    #        print("root=",root)
    #        print("dirs=",dirs)
    #        print("files=",files)

    #####volume optimization for simple cubic systems########
    if 'INFO_VOL' in os.listdir('.'):
      self.volumeOpt = True
      #  print("self.volumeOpt=",self.volumeOpt)
      with open('INFO_VOL') as g:
          while 1:
            s = g.readline()
            if not s: break
            s = s.strip()
    else:
      for files in os.listdir('.'):
      #  print("files=",files)
      #  backend.addValue("x_exciting_dummy2", files)
        if files[0:7] == 'rundir-':
          self.volumeCubOptIndex+=1
    os.chdir(curDir)
    #  curDir = os.getcwd()
    #  dirPath = self.mainFileUri[0:-9]
    #  os.chdir(dirPath)
    #  os.chdir('../')
    #  i = 0
    #  for files in os.listdir('.'):
    #    if files[0:7] == 'rundir-':
    #      i+=1
    #    if i>1:
    #      self.volumeOpt = True
    #      optGindex = backend.openSection("section_method")
    #      print("optGindex=",optGindex)
    #      backend.addValue("x_exciting_volume_optimization", self.volumeOpt)
    #      backend.closeSection("section_method", optGindex)
    #  os.chdir(curDir)

  def onOpen_section_sampling_method(self, backend, gIndex, section):
    self.secSamplingMethodIndex = gIndex
    backend.addValue("sampling_method", "geometry_optimization")
    #  print("self.secSamplingMethodIndex=",self.secSamplingMethodIndex)
    self.samplingMethod = "geometry_optimization"
    #  print("self.samplingMethod=",self.samplingMethod)

  def onOpen_section_system(self, backend, gIndex, section):
    self.secSystemIndex = gIndex
    curDir = os.getcwd()
    mainFile = self.parser.fIn.fIn.name           #####exciting sbt -> zip file or from NOMAD URI nmd:// ####     YES ?????????? sure???? check first    OKOKOKKO
    self.mainFilePath = os.path.dirname(mainFile)           #####exciting sbt -> zip file####     YES ?????????? sure???? check first    OKOKOKKO
    #  self.mainFilePath = self.mainFileUri[0:-9]     #####exciting LOCAL HOME#######   YES                      OKOKOKOK
    #  print("self.mainFilePath===",self.mainFilePath)
    os.chdir(self.mainFilePath)
    #  print("listdir=",os.listdir('.'))
    if 'str.out' in os.listdir('.'):
      self.clathrates = True
      #  print("clathrate_vero===",self.clathrates)
      backend.addValue("x_exciting_clathrates", True)
      clathrate_labels = []
      clathrate_coordinates = []
      with open('str.out') as g:
          while 1:
            s = g.readline()
            if not s: break
            s = s.strip()
            s = s.split()
            if len(s) == 4:
              #  print("s=",s[0],float(s[0]))
              clathrate_coordinates.append([float(s[0]),float(s[1]),float(s[2])])
              clathrate_labels.append(s[3])
      #  print("clathrate_coordinates=",clathrate_coordinates)
      backend.addArrayValues("x_exciting_clathrates_atom_coordinates", np.asarray(clathrate_coordinates))
      backend.addValue("x_exciting_clathrates_atom_labels", clathrate_labels)
    else:
      #  print("clathrate_falso===",self.clathrates)
      backend.addValue("x_exciting_clathrates", False)
      #  backend.addArrayValues("x_exciting_clathrates_atom_coordinates", np.array(clathrate_coordinates))
      #  backend.addValue("x_exciting_clathrates_atom_labels", clathrate_labels)
    os.chdir(curDir)

  def onOpen_section_single_configuration_calculation(self, backend, gIndex, section):
    if self.secSingleConfIndex is None:
      self.secSingleConfIndex = gIndex
    self.frameSequence.append(gIndex)

  def onOpen_section_method(self, backend, gIndex, section):
    if self.secMethodIndex is None:
      self.secMethodIndex = gIndex

  def onOpen_x_exciting_section_geometry_optimization(self, backend, gIndex, section):
    #    """Trigger called when x_abinit_section_dataset is opened.
    #    """
    self.samplingGIndex = backend.openSection("section_sampling_method")

  def onClose_section_run(self, backend, gIndex, section):
    self.secRunIndex = gIndex
    #  backend.addValue("x_exciting_dummy", self.volumeOptIndex)
    ########### VOLUME OPTIMIZATION #########################
    #    curDir = os.getcwd()
    ##    mainFile = self.parser.fIn.fIn.name           #####exciting sbt -> zip file####     YES ?????????? sure???? check first    OKOKOKKO
    ##    dirPath = os.path.dirname(mainFile)           #####exciting sbt -> zip file####     YES ?????????? sure???? check first    OKOKOKKO
    #    dirPath = self.mainFileUri[0:-9]     #####exciting LOCAL HOME or from NOMAD URI nmd://  #######   YES                      OKOKOKOK
    #    os.chdir(dirPath)
    #    os.chdir('../')
    #    self.volumeOptIndex = 0
    #    for files in os.listdir('.'):
    #      if files[0:7] == 'rundir-':
    #        self.volumeOptIndex+=1
    ######################independent from above#################
    if self.volumeCubOptIndex>1:
      self.volumeCubOpt = True
      optGindex = backend.openSection("section_method")
      backend.addValue("x_exciting_volume_optimization", self.volumeCubOpt)
      backend.closeSection("section_method", optGindex)
    #  os.chdir(curDir)

    ####################################TEST############################
    mainFile = self.parser.fIn.fIn.name
    dirPath = os.path.dirname(self.parser.fIn.name)
    #  print("dirPath===",dirPath)
    gw_File = os.path.join(dirPath, "GW_INFO.OUT")
    gwFile = os.path.join(dirPath, "GWINFO.OUT")
    xsFile = os.path.join(dirPath, "INFOXS.OUT")
    for gFile in [gw_File, gwFile]:
      if os.path.exists(gFile):
        gwParser = exciting_parser_gw.GWParser()
        gwParser.parseGW(gFile, backend,
                         dftMethodSectionGindex = self.secMethodIndex,
                         dftSingleConfigurationGindex = self.secSingleConfIndex,
                         xcName = self.xcName,
                         unitCellVol = self.unit_cell_vol,
                         gmaxvr = self.gmaxvr)

        subParser = AncillaryParser(
            fileDescription = exciting_parser_gw.buildGWMatchers(),
            parser = self.parser,
            cachingLevelForMetaName = exciting_parser_gw.get_cachingLevelForMetaName(self.metaInfoEnv, CachingLevel.PreOpenedIgnore),
            superContext = gwParser)
        with open(gFile) as fIn:
            subParser.parseFile(fIn)
        break

    if os.path.exists(xsFile):
      excNum = []
      excEn = []
      excBindEn = []
      osclStr = []
      transCoeff = []
      epsEn = []
      sigmaEn = []
      epsilon = []
      sigma = []
      lossEn = []
      loss = []
      lossDummy = []
      xstype = "TDDFT"
      self.xstype = "TDDFT"
      fromeV = unit_conversion.convert_unit_function("eV", "J")
      #  print("BSE!!!")
      #  if os.path.exists(inputgwFile):
      inputXSFile = os.path.join(dirPath, "input.xml")
      self.XSSetGIndex = backend.openSection("section_method")
      try:  # tmk:
        with open(inputXSFile) as f:
          exciting_parser_XS_input.parseInput(f, backend, self.rgkmax)
      except FileNotFoundError:
        logger.warning("File not found: {}" .format(inputXSFile))
      except Exception as err:
        logger.error("Exception on {}" .format(__file__), exc_info=err)
        #    xstype = section["x_exciting_xs_type"]
        #    print("xstype===",xstype)
        #    print("xsType===",exciting_parser_XS_input.InputHandler.self.xsType)
        #  backend.addValue('x_exciting_electronic_structure_method', "BSE")   ########!!!!!!!!!!! So far, BSE is not a valid value for electronic_structure_method. It must be added! #########
        #    teste = section["x_exciting_xs_tetra"]
        #    print("teste===",teste)
        #  backend.addValue('x_exciting_xs_starting_point', self.xcName)
        #  teste = section["x_exciting_xs_tetra"]
        #  print("teste===",teste)
      if self.secMethodIndex is not None:
        m2mGindex = backend.openNonOverlappingSection("section_method_to_method_refs")
        backend.addValue("method_to_method_ref", self.secMethodIndex)
        backend.addValue("method_to_method_kind", "starting_point")
        backend.closeNonOverlappingSection("section_method_to_method_refs")

        #  xstype = section["x_exciting_xs_type"]
        #  print("xstype===",xstype)
        for files in os.listdir(dirPath):
            if files[0:11] == "EXCITON_BSE":
                xstype = "BSE"
                self.xstype = "BSE"
                dummyBse = files[11:13]
                #  self.tensorComp = files[-6:-4]
                self.tensorComp.append(files[-6:-4])

                if dummyBse == 'si':
                    self.bsetype = 'singlet'
                    #  name = "EXCITON_BSE" + self.bsetype
                elif dummyBse == 'tr':
                    self.bsetype = 'triplet'
                elif dummyBse == 'RP':
                    self.bsetype = 'RPA'
                elif dummyBse == 'IP':
                    self.bsetype = 'IP'
                else:
                    self.bsetype = 'UNKNOWN'
                name = 'EXCITON_BSE%s_SCR' % self.bsetype
                #  nameEps = "EPSILON_BSE" + self.bsetype + '_SCR'
                if files[len(name):len(name)+4] == 'full':
                    self.screentype = 'full'
                elif files[len(name):len(name)+4] == 'diag':
                    self.screentype = 'diag'
                elif files[len(name):len(name)+4] == 'noin':
                    self.screentype = 'noinvdiag'
                elif files[len(name):len(name)+4] == 'long':
                    self.screentype = 'longrange'
          #  else:
          #    xstype = "TDDFT"
          #  print("xstype===",xstype)
          #  teste = section["x_exciting_xs_tetra"]
          #  print("teste===",teste)
        if xstype == "BSE":
            numberOfComponents = len(self.tensorComp)
            #  backend.addValue("x_exciting_xs_screening_type", self.screentype)
            backend.addValue("x_exciting_xs_bse_type", self.bsetype)
      #  print("===", self.tensorComp[0])
      #  with open(inputXSFile) as f:
      #    exciting_parser_XS_input.parseInput(f, backend, self.rgkmax)
      #    teste = section["x_exciting_xs_tetra"]
      #    print("teste===",teste)
      #  teste = section["x_exciting_xs_tetra"]
      #  print("teste===",teste)
      backend.closeSection("section_method",self.XSSetGIndex)
      #  teste = section["x_exciting_xs_tetra"]
      #  print("teste===",teste)
      #  numberOfComponents = len(self.tensorComp)
      #  print("xstype=====",xstype)
      backend.openNonOverlappingSection("section_single_configuration_calculation")
      if self.secSingleConfIndex is not None:
          backend.openNonOverlappingSection("section_calculation_to_calculation_refs")
          backend.addValue("calculation_to_calculation_ref", self.secSingleConfIndex)
          backend.addValue("calculation_to_calculation_kind", "starting_point")
          backend.closeNonOverlappingSection("section_calculation_to_calculation_refs")

      if xstype == "BSE":
          for i in range(numberOfComponents):
              excNum.append([])
              excEn.append([])
              epsEn.append([])
              sigmaEn.append([])
              excBindEn.append([])
              osclStr.append([])
              transCoeff.append([[],[]])
              epsilon.append([[],[]])
              sigma.append([[],[]])
              lossEn.append([])
              loss.append([[],[]])
            #  lossDummy.append([])

              try:
                outputXSFile = os.path.join(dirPath, "EXCITON_BSE" + self.bsetype + '_SCR' + self.screentype + "_OC" + self.tensorComp[i] + ".OUT")
                outputEpsFile = os.path.join(dirPath, "EPSILON_BSE" + self.bsetype + '_SCR' + self.screentype + "_OC" + self.tensorComp[i] + ".OUT")
                outputSigmaFile = os.path.join(dirPath, "SIGMA_BSE" + self.bsetype + '_SCR' + self.screentype + "_OC" + self.tensorComp[i] + ".OUT")
                outputLossFile = os.path.join(dirPath, "LOSS_BSE" + self.bsetype + '_SCR' + self.screentype + "_OC" + self.tensorComp[i] + ".OUT")
                # - - - - -
                try:
                  with open(outputXSFile) as g:
                      xsParser = exciting_parser_xs.XSParser()
                      xsParser.parseExciton(outputXSFile, backend, excNum, excEn, excBindEn, osclStr, transCoeff) #, dftMethodSectionGindex = self.secMethodIndex,
                except FileNotFoundError:
                  logger.warning("File not found: {}" .format(outputXSFile))
                except Exception as err:
                  logger.error("Exception on {}" .format(__file__), exc_info=err)
                # - - - - -
                try:
                  with open(outputEpsFile) as g:
                      epsParser = exciting_parser_eps.EPSParser()
                      epsParser.parseEpsilon(outputEpsFile, backend, epsEn, epsilon) #, dftMethodSectionGindex = self.secMethodIndex,
                except FileNotFoundError:
                  logger.warning("File not found: {}" .format(outputEpsFile))
                except Exception as err:
                  logger.error("Exception on Exciting subparser", exc_info=err)
                # - - - - -
                try:
                  with open(outputSigmaFile) as g:
                      sigmaParser = exciting_parser_eps.EPSParser()
                      sigmaParser.parseEpsilon(outputSigmaFile, backend, sigmaEn, sigma) #, dftMethodSectionGindex = self.secMethodIndex,
                except FileNotFoundError:
                  logger.warning("File not found: {}" .format(outputSigmaFile))
                except Exception as err:
                  logger.error("Exception on {}" .format(__file__), exc_info=err)
              except Exception as err:
                logger.error('Exception while processing further outfiles.')


              # with open(outputLossFile) as g:
              #     lossParser = exciting_parser_eps.EPSParser()
              #     lossParser.parseEpsilon(outputLossFile, backend, lossEn, loss) #, dftMethodSectionGindex = self.secMethodIndex,
              #     dftSingleConfigurationGindex = self.secSingleConfIndex)
          backend.addValue("x_exciting_xs_bse_number_of_components",numberOfComponents)
          backend.addValue("x_exciting_xs_bse_number_of_excitons",len(excNum))
          backend.addValue("x_exciting_xs_bse_number_of_energy_points",len(epsEn))
          backend.addValue("x_exciting_xs_bse_exciton_energies",excEn)
          backend.addValue("x_exciting_xs_bse_exciton_binding_energies",excBindEn)
          backend.addValue("x_exciting_xs_bse_exciton_oscillator_strength",osclStr)
          backend.addValue("x_exciting_xs_bse_exciton_amplitude_re",transCoeff[0])
          backend.addValue("x_exciting_xs_bse_exciton_amplitude_im",transCoeff[1])
          backend.addValue("x_exciting_xs_bse_epsilon_energies",epsEn)
          backend.addValue("x_exciting_xs_bse_epsilon_re",epsilon[0])
          backend.addValue("x_exciting_xs_bse_epsilon_im",epsilon[1])
          backend.addValue("x_exciting_xs_bse_sigma_energies",sigmaEn)
          backend.addValue("x_exciting_xs_bse_sigma_re",sigma[0])
          backend.addValue("x_exciting_xs_bse_sigma_im",sigma[1])
          backend.addValue("x_exciting_xs_bse_loss_energies",lossEn)
          backend.addValue("x_exciting_xs_bse_loss",loss[0])

      if xstype == "TDDFT":
          dielFunctEne = []
          dielFunctLoc=[[],[]]
          dielFunctNoLoc=[[],[]]
          dielTensNoSym = []
          dielTensSym = []
          lossFunctionLoc = []
          lossFunctionNoLoc = []
          qCartesian = []
          QFile = os.path.join(dirPath, "QPOINTS.OUT")
          qLattice = []
          qPlusG = []
          qPlusGCartesian = []
          qPlusGLattice = []
          qpointNumber = 0
          sigmaLoc = [[],[]]
          sigmaNoLoc = [[],[]]
          tensorComp=[]

          for files in os.listdir(dirPath):
              if files[0:9] == "SIGMA_NLF":
                  tensorComp.append(files[-13:-11])
                  #  print("tensorComp===",tensorComp)
          if self.xsTetra and self.xsAC and not self.xsNAR:
              ext = "TET_AC_NAR"
          elif not self.xsTetra and self.xsAC and not self.xsNAR:
              ext = "AC_NAR"
          elif not self.xsTetra and not self.xsAC and not self.xsNAR:
              ext = "NAR"
          elif self.xsTetra and self.xsAC and self.xsNAR:
              ext = "TET_AC"
          elif self.xsTetra and not self.xsAC and self.xsNAR:
              ext = "TET"
          elif self.xsTetra and not self.xsAC and not self.xsNAR:
              ext = "TET_NAR"
          else:
              ext=""
          try:
            with open(QFile) as g:
                while 1:
                    s = g.readline()
                    if not s: break
                    s = s.strip()
                    s = s.split()
                    if not is_number(s[1]):
                        qpointNumber = int(s[0] )
                    else:
                        qPlusGCartesian.append([])
                        qPlusGLattice.append([])
                        qLattice.append([float(s[1]),float(s[2]),float(s[3])])
                        qCartesian.append([float(s[1]),float(s[2]),float(s[3])])
                        qPlusG.append(int(s[7]))
                if self.xsTetra and self.xsAC and not self.xsNAR:
                    ext = "TET_AC_NAR"
                elif not self.xsTetra and self.xsAC and not self.xsNAR:
                    ext = "AC_NAR"
                elif not self.xsTetra and not self.xsAC and not self.xsNAR:
                    ext = "NAR"
                elif self.xsTetra and self.xsAC and self.xsNAR:
                    ext = "TET_AC"
                elif self.xsTetra and not self.xsAC and self.xsNAR:
                    ext = "TET"
                elif self.xsTetra and not self.xsAC and not self.xsNAR:
                    ext = "TET_NAR"
                else:
                    ext=""
          except FileNotFoundError:
            logger.warning("File not found: {}" .format(QFile))
          except Exception as err:
            logger.error("Exception on {}" .format(__file__), exc_info=err)
          # ---- QFile closing
          #  xstype = "BSE"
          #  self.xstype = "BSE"
          #  dummyBse = files[11:13]
          #  self.tensorComp = files[-6:-4]
          # WARNING: the following entries don't exist in Exciting Metainfo!!
          #backend.addValue("x_exciting_xs_tddft_number_of_optical_components",len(tensorComp))
          #backend.addValue("x_exciting_xs_tddft_optical_component",tensorComp)
          #backend.addValue("s",qpointNumber)

          for i in range(qpointNumber):
              dielTensSym.append([[],[]])
              dielTensNoSym.append([[],[]])
              dielFunctLoc[0].append([])
              dielFunctLoc[1].append([])
              dielFunctNoLoc[0].append([])
              dielFunctNoLoc[1].append([])
              lossFunctionLoc.append([])
              lossFunctionNoLoc.append([])
              dielFunctEne.append([])
              sigmaLoc[0].append([])
              sigmaLoc[1].append([])
              sigmaNoLoc[0].append([])
              sigmaNoLoc[1].append([])
              if i < 10:
                  qExt00 = '_QMT00'+ str(i+1)
                  qPlusGFile = os.path.join(dirPath, 'GQPOINTS' + qExt00 + '.OUT')
                  DielNoSymFile = os.path.join(dirPath, 'DIELTENS0_NOSYM' + qExt00 + '.OUT')
                  DielSymFile = os.path.join(dirPath, 'DIELTENS0' + qExt00 + '.OUT')
                  with open(qPlusGFile) as g:
                      while 1:
                          s = g.readline()
                          if not s: break
                          s = s.strip()
                          s = s.split()
                          if is_number(s[1]):
                              qPlusGLattice[i].append([float(s[1]),float(s[2]),float(s[3])])
                              qPlusGCartesian[i].append([float(s[4]),float(s[5]),float(s[6])])
                  with open(DielNoSymFile) as g:
                      while 1:
                          s = g.readline()
                          if not s: break
                          s = s.strip()
                          s = s.split()
                          if s and is_number(s[0]):
                              dielTensNoSym[i][0].append([float(s[0]),float(s[1]),float(s[2])])
                              dielTensNoSym[i][1].append([float(s[3]),float(s[4]),float(s[5])])
                  with open(DielSymFile) as g:
                      while 1:
                          s = g.readline()
                          if not s: break
                          s = s.strip()
                          s = s.split()
                          if s and is_number(s[0]):
                              dielTensSym[i][0].append([float(s[0]),float(s[1]),float(s[2])])
                              dielTensSym[i][1].append([float(s[3]),float(s[4]),float(s[5])])
                  for j in range(len(tensorComp)):
                      dielFunctLoc[0][-1].append([])
                      dielFunctLoc[1][-1].append([])
                      dielFunctNoLoc[0][-1].append([])
                      dielFunctNoLoc[1][-1].append([])
                      dielFunctEne[-1].append([])
                      lossFunctionLoc[-1].append([])
                      lossFunctionNoLoc[-1].append([])
                      sigmaLoc[0][-1].append([])
                      sigmaLoc[1][-1].append([])
                      sigmaNoLoc[0][-1].append([])
                      sigmaNoLoc[1][-1].append([])
                      #  print("j===",j)
                      #  print("ext===",ext)
                      #  print("self.tddftKernel===",self.tddftKernel)
                      #  print("qExt00===",qExt00)
                      #  print("tutto===",'EPSILON_' + ext + 'FXC' + self.tddftKernel[0] + '_OC' + tensorComp[j] + qExt00 + '.OUT')

                      if self.tddftKernel is not None:
                        ########
                        try:
                          epsilonLocalField = os.path.join(dirPath, 'EPSILON_' + ext + 'FXC' + self.tddftKernel[0] + '_OC' + tensorComp[j] + qExt00 + '.OUT')
                          with open(epsilonLocalField) as g:
                            while 1:
                              s = g.readline()
                              if not s: break
                              s = s.strip()
                              s = s.split()
                              ene, epsRe, epsIm = fromeV(float(s[0])),float(s[1]),float(s[2])
                              dielFunctLoc[0][-1][-1].append(epsRe)
                              dielFunctLoc[1][-1][-1].append(epsIm)
                              dielFunctEne[-1][-1].append(ene)
                          backend.addValue("x_exciting_xs_tddft_number_of_epsilon_values",len(dielFunctEne[0][0]))
                          backend.addValue("x_exciting_xs_tddft_epsilon_energies",dielFunctEne[0][0])
                          backend.addValue("x_exciting_xs_tddft_dielectric_function_local_field",dielFunctLoc)
                        except IOError:
                          logger.error("File not processable: %s" % (epsilonLocalField))

                        #########
                        try:
                          epsilonNoLocalField = os.path.join(dirPath, 'EPSILON_' + ext + 'NLF_' + 'FXC' + self.tddftKernel[0] + '_OC' + tensorComp[j] + qExt00 + '.OUT')
                          with open(epsilonNoLocalField) as g:
                            while 1:
                              s = g.readline()
                              if not s: break
                              s = s.strip()
                              s = s.split()
                              epsRe, epsIm = float(s[1]),float(s[2])
                              dielFunctNoLoc[0][-1][-1].append(epsRe)
                              dielFunctNoLoc[1][-1][-1].append(epsIm)
                          backend.addValue("x_exciting_xs_tddft_dielectric_function_no_local_field",dielFunctNoLoc)
                        except IOError:
                          logging.error("File not processable: %s" % (epsilonNoLocalField))

                        #########
                        try:
                          lossFunctionLocalFieldFile = os.path.join(dirPath, 'LOSS_' + ext + 'FXC' + self.tddftKernel[0] + '_OC' + tensorComp[j] + qExt00 + '.OUT')
                          with open(lossFunctionLocalFieldFile) as g:
                            while 1:
                              s = g.readline()
                              if not s: break
                              s = s.strip()
                              s = s.split()
                              #print("s===",s)
                              if s and is_number(s[0]):
                                loss = float(s[1])
                                lossFunctionLoc[-1][-1].append(loss)
                          backend.addValue("x_exciting_xs_tddft_loss_function_local_field",lossFunctionLoc)
                        except IOError:
                          logger.error("File not processable: %s" % (lossFunctionLocalFieldFile))

                        #########
                        try:
                          lossFunctionNoLocalFieldFile = os.path.join(dirPath, 'LOSS_' + ext + 'NLF_' + 'FXC' + self.tddftKernel[0] + '_OC' + tensorComp[j] + qExt00 + '.OUT')
                          with open(lossFunctionNoLocalFieldFile) as g:
                            while 1:
                              s = g.readline()
                              if not s: break
                              s = s.strip()
                              s = s.split()
                              # print("s===",s)
                              if s and is_number(s[0]):
                                loss = float(s[1])
                                lossFunctionNoLoc[-1][-1].append(loss)
                          backend.addValue("x_exciting_xs_tddft_loss_function_no_local_field",lossFunctionNoLoc)
                        except IOError:
                          logger.error("File not processable: %s" % (lossFunctionNoLocalFieldFile))

                        #########
                        sigmaLocalFieldFile = os.path.join(dirPath, 'SIGMA_' + ext + 'FXC' + self.tddftKernel[0] + '_OC' + tensorComp[j] + qExt00 + '.OUT')
                        try:
                          with open(sigmaLocalFieldFile) as g:
                            while 1:
                              s = g.readline()
                              if not s: break
                              s = s.strip()
                              s = s.split()
                              if s and is_number(s[0]):
                                sigmaRe, sigmaIm = float(s[1]),float(s[2])
                                sigmaLoc[0][-1][-1].append(sigmaRe)
                                sigmaLoc[1][-1][-1].append(sigmaIm)
                          backend.addValue("x_exciting_xs_tddft_sigma_local_field",sigmaLoc)
                        except IOError:
                          logger.error("File not processable: %s" % (sigmaLocalFieldFile))
                        #
                        #########
                        sigmaNoLocalFieldFile = os.path.join(dirPath, 'SIGMA_' + ext + 'NLF_' + 'FXC' + self.tddftKernel[0] + '_OC' + tensorComp[j] + qExt00 + '.OUT')
                        try:
                          with open(sigmaNoLocalFieldFile) as g:
                            while 1:
                              s = g.readline()
                              if not s: break
                              s = s.strip()
                              s = s.split()
                              if s and is_number(s[0]):
                                sigmaRe, sigmaIm = float(s[1]),float(s[2])
                                sigmaNoLoc[0][-1][-1].append(sigmaRe)
                                sigmaNoLoc[1][-1][-1].append(sigmaIm)
                          backend.addValue("x_exciting_xs_tddft_sigma_no_local_field",sigmaNoLoc)
                        except IOError:
                          logger.error("File not processable: %s" % (sigmaLocalFieldFile))

          #  dielFunctEne[-1][-1].append(ene)
          #  if s and is_number(s[0]):
          #    dielTensSym[i][0].append([float(s[0]),float(s[1]),float(s[2])])
          #    dielTensSym[i][1].append([float(s[3]),float(s[4]),float(s[5])])
          # print("loss===",lossFunctionLoc)
          # print("ext===",ext)
          # print("dielTensSym===",dielTensSym)
          # print("dielTensNoSym===",dielTensNoSym)
          #    qPlusGLattice[i].append([float(s[1]),float(s[2]),float(s[3])])
          #    qPlusGCartesian[i].append([float(s[4]),float(s[5]),float(s[6])])
          #WARNING: the following entries don't exist in the Exciting Metainfo!!
          #backend.addValue("x_exciting_xs_tddft_dielectric_tensor_sym",dielTensSym)
          #backend.addValue("x_exciting_xs_tddft_dielectric_tensor_no_sym",dielTensNoSym)

      backend.closeNonOverlappingSection("section_single_configuration_calculation")


  def onClose_x_exciting_section_geometry_optimization(self, backend, gIndex, section):
    """Trigger called when x_abinit_section_dataset is closed.
    """
    if len(self.frameSequence) > 1:
      frameGIndex = backend.openSection("section_frame_sequence")
      backend.addValue("geometry_optimization_converged", True)
      backend.closeSection("section_frame_sequence", frameGIndex)
    backend.closeSection("section_sampling_method", self.samplingGIndex)

  def onClose_section_frame_sequence(self, backend, gIndex, section):
    """Trigger called when section_framce_sequence is closed.
    """
    backend.addValue("number_of_frames_in_sequence", len(self.frameSequence))
    backend.addArrayValues("frame_sequence_local_frames_ref", np.array(self.frameSequence))
    backend.addValue("frame_sequence_to_sampling_ref", self.samplingGIndex)

    #  print("self.samplingMethod=",self.samplingMethod)
    if self.samplingMethod == "geometry_optimization":
    #  gix = backend.openSection("section_sampling_method")
    #  backend.addValue("XC_functional_name", xcName)
    #  backend.closeSection("section_sampling_method", gix)
    #  geometryForceThreshold = section["x_exciting_geometry_optimization_threshold_force"]
    #  print("geometryForceThreshold=",self.geometryForceThreshold)
        gi = backend.openSection("section_sampling_method")
        backend.addValue("geometry_optimization_threshold_force", self.geometryForceThreshold)
        backend.closeSection("section_sampling_method", gi)
    else:
        pass

  def onClose_x_exciting_section_lattice_vectors(self, backend, gIndex, section):
    latticeX = section["x_exciting_geometry_lattice_vector_x"]
    latticeY = section["x_exciting_geometry_lattice_vector_y"]
    latticeZ = section["x_exciting_geometry_lattice_vector_z"]
    cell = [[latticeX[0],latticeY[0],latticeZ[0]],
            [latticeX[1],latticeY[1],latticeZ[1]],
            [latticeX[2],latticeY[2],latticeZ[2]]]
    self.sim_cell = cell
    backend.addValue("simulation_cell", cell)

  def onClose_x_exciting_section_reciprocal_lattice_vectors(self, backend, gIndex, section):
    recLatticeX = section["x_exciting_geometry_reciprocal_lattice_vector_x"]
    recLatticeY = section["x_exciting_geometry_reciprocal_lattice_vector_y"]
    recLatticeZ = section["x_exciting_geometry_reciprocal_lattice_vector_z"]
    recCell = [[recLatticeX[0],recLatticeY[0],recLatticeZ[0]],
            [recLatticeX[1],recLatticeY[1],recLatticeZ[1]],
            [recLatticeX[2],recLatticeY[2],recLatticeZ[2]]]
    backend.addValue("x_exciting_simulation_reciprocal_cell", recCell)

  def onClose_x_exciting_section_xc(self, backend, gIndex, section):
    xcNr = section["x_exciting_xc_functional"][0]
    xc_internal_map = {
        2: ['LDA_C_PZ', 'LDA_X_PZ'],
        3: ['LDA_C_PW', 'LDA_X_PZ'],
        4: ['LDA_C_XALPHA'],
        5: ['LDA_C_VBH'],
        20: ['GGA_C_PBE', 'GGA_X_PBE'],
        21: ['GGA_C_PBE', 'GGA_X_PBE_R'],
        22: ['GGA_C_PBE_SOL', 'GGA_X_PBE_SOL'],
        26: ['GGA_C_PBE', 'GGA_X_WC'],
        30: ['GGA_C_AM05', 'GGA_C_AM05'],
        300: ['GGA_C_BGCP', 'GGA_X_PBE'],
        406: ['HYB_GGA_XC_PBEH']
        }
    if xcNr == 100:
        dirPath = os.path.dirname(self.parser.fIn.name)
        inputGSFile = os.path.join(dirPath, "input.xml")
        try:
          with open(inputGSFile) as f:
            exciting_parser_GS_input.parseInput(f, backend)
        except FileNotFoundError:
          logger.warning("File not found: {}" .format(inputGSFile))
        except Exception as err:
          logger.error("Exception while processing file {}" .format(inputGSFile), exc_info=err)
    else:
        for xcName in xc_internal_map[xcNr]:
          self.xcName = xcName
          gi = backend.openSection("section_XC_functionals")
          backend.addValue("XC_functional_name", xcName)
          backend.closeSection("section_XC_functionals", gi)

  def onClose_section_single_configuration_calculation(self, backend, gIndex, section):
    # Determine the Fermi energy for this calculation.
    energy_reference_fermi = None
    try:
        energy_reference_fermi = backend.superBackend.get_value('x_exciting_fermi_energy', g_index=gIndex).m
    except KeyError:
        try:
            energy_reference_fermi = backend.superBackend.get_value('gw_fermi_energy', g_index=gIndex).m
        except KeyError:
            pass
    if energy_reference_fermi is not None:
        backend.addArrayValues('energy_reference_fermi', [energy_reference_fermi, energy_reference_fermi])  # always two spin channels

    # logger.error("BASE onClose_section_single_configuration_calculation")
    backend.addValue('single_configuration_to_calculation_method_ref', self.secMethodIndex)
    backend.addValue('single_configuration_calculation_to_system_ref', self.secSystemIndex)
    # print("self.samplingMethod=",self.samplingMethod)

    ####################VOLUME TEST BEGIN################################
    ext_uri = []
    # backend.addValue("x_exciting_dummy=",self.volumeOptIndex)
    # backend.addValue("x_exciting_dummy", self.volumeOptIndex)
    if self.volumeCubOptIndex > 1:
      for j in range(1, self.volumeCubOptIndex):
        if (j<10):
          ext_uri.append(self.mainFilePath[0:-9] + 'rundir-0' + str(j) + '/INFO.OUT')
          # backend.addValue("x_exciting_dummy2", ext_uri[-1])
          # print("ext_uri===",ext_uri)
        else:
          ext_uri.append(self.mainFilePath[0:-9] + 'rundir-' + str(j) + '/INFO.OUT')
          # backend.addValue("x_exciting_dummy2", ext_uri[-1])
          # print("ext_uri===",ext_uri)
      # backend.addArrayValues("x_exciting_dummy2", np.asarray(ext_uri))
    for ref in ext_uri:
      refGindex = backend.openSection("section_calculation_to_calculation_refs")
      backend.addValue("calculation_to_calculation_external_url", ref)
      backend.addValue("calculation_to_calculation_kind", "source_calculation")
      backend.closeSection("section_calculation_to_calculation_refs", refGindex)

    ####################VOLUME TEST END############################

    if self.samplingMethod == "geometry_optimization":
      ivalue = section["x_exciting_geometry_optimization_threshold_force"]
      if ivalue is None:
        # then use default value
        self.geometryForceThreshold = self.geometryForceThreshold
        logger.warning("Found suspicious value for geometry optimization "
          "threshold force. Hint: inspect INFO.OUT, is it complete?")
      else:
        try:
          self.geometryForceThreshold = ivalue[0]
        except:
          raise
    forceX = section["x_exciting_geometry_atom_forces_x"]
    if forceX:
      forceY = section["x_exciting_geometry_atom_forces_y"]
      forceZ = section["x_exciting_geometry_atom_forces_z"]
      atoms = len(forceX)
      atom_geometry_forces = []
      for i in range(0,atoms):
          atom_geometry_forces.append([forceX[i],forceY[i],forceZ[i]])
      backend.addValue("atom_forces",atom_geometry_forces)
      # print("geometryForceThreshold=",geometryForceThreshold)
      # backend.addValue("geometry_optimization_threshold_force", geometryForceThreshold)
    # else:
    #   pass
    #
    ##############TO DO. FIX FORCES#####################
    #    forceX = section["x_exciting_atom_forces_x"]
    #    if forceX:
    #      forceY = section["x_exciting_atom_forces_y"]
    #      forceZ = section["x_exciting_atom_forces_z"]
    #      print("forceX===",forceX)
    #      print("forceY===",forceY)
    #      print("forceZ===",forceZ)
    #      forceCoreX = section["x_exciting_atom_core_forces_x"]
    #      forceCoreY = section["x_exciting_atom_core_forces_y"]
    #      forceCoreZ = section["x_exciting_atom_core_forces_z"]
    #      print("forceCoreX===",forceCoreX)
    ##      print("forceCoreY===",forceCoreY)
    #      print("forceCoreZ===",forceCoreZ)
    #      forceIBSX = section["x_exciting_atom_IBS_forces_x"]
    #      forceIBSY = section["x_exciting_atom_IBS_forces_y"]
    #      forceIBSZ = section["x_exciting_atom_IBS_forces_z"]
    #      print("forceIBSX===",forceIBSX)
    #      print("forceIBSY===",forceIBSY)
    #      print("forceIBSZ===",forceIBSZ)
    #      forceHFX = section["x_exciting_atom_HF_forces_x"]
    #      forceHFY = section["x_exciting_atom_HF_forces_y"]
    #      forceHFZ = section["x_exciting_atom_HF_forces_z"]
    #      fConv = convert_unit_function("hartree/bohr", "N")
    #      atoms = len(forceX)
    #      atom_forces = []
    #      atom_core_forces = []
    #      atom_IBS_forces = []
    #      atom_HF_forces = []
    #      for i in range(0,atoms):
    #        print("atoms===",atoms)
    #        print("i===",i)
    #        print("atom_forces===",atom_forces)
    #        print("forceX[i]===",forceX[i])
    #        print("forceY[i]===",forceY[i])
    #        print("forceZ[i]===",forceZ[i])
    #        print("forceCoreX[i]===",forceCoreX[i])
    #        print("forceCoreY[i]===",forceCoreY[i])
    #        print("forceCoreZ[i]===",forceCoreZ[i])
    #        print("forceIBSX[i]===",forceIBSX[i])
    #        print("forceIBSY[i]===",forceIBSY[i])
    #        print("forceIBSZ[i]===",forceIBSZ[i])
    #        atom_forces.append([fConv(forceX[i]),fConv(forceY[i]),fConv(forceZ[i])])
    #        atom_core_forces.append([fConv(forceCoreX[i]),fConv(forceCoreY[i]),fConv(forceCoreZ[i])])
    #        atom_IBS_forces.append([fConv(forceIBSX[i]),fConv(forceIBSY[i]),fConv(forceIBSZ[i])])
    #        atom_HF_forces.append([fConv(forceHFX[i]),fConv(forceHFY[i]),fConv(forceHFZ[i])])
    #      backend.addValue("atom_forces",atom_forces)
    #      backend.addValue("x_exciting_atom_core_forces",atom_core_forces)
    #      backend.addValue("x_exciting_atom_IBS_forces",atom_IBS_forces)
    #      backend.addValue("x_exciting_atom_HF_forces",atom_HF_forces)
    #    print("atom_forces=",atom_forces)
    #
    dirPath = os.path.dirname(self.parser.fIn.name)
    dosFile = os.path.join(dirPath, "dos.xml")
    bandFile = os.path.join(dirPath, "bandstructure.xml")
    fermiSurfFile = os.path.join(dirPath, "FERMISURF.bxsf")
    eigvalFile = os.path.join(dirPath, "EIGVAL.OUT")
    #    logger.error("done BASE onClose_section_single_configuration_calculation")

    try:
      with open(dosFile) as f:
        exciting_parser_dos.parseDos(f, backend, self.spinTreat, self.unit_cell_vol, energy_reference_fermi)
    except FileNotFoundError:
      logger.warning("File not found: {}" .format(dosFile))
    except Exception as err:
      logger.error("Exception while processing file {}" .format(dosFile), exc_info=err)

    if os.path.exists(bandFile):
      with open(bandFile) as g:
        exciting_parser_bandstructure.parseBand(g, backend, self.spinTreat)
    if os.path.exists(eigvalFile):
      eigvalGIndex = backend.openSection("section_eigenvalues")
      with open(eigvalFile) as g:
          eigvalKpoint=[]
          eigvalVal=[]
          eigvalOcc=[]
          eigvalValSpin = [[],[]]
          eigvalOccSpin = [[],[]]
          fromH = unit_conversion.convert_unit_function("hartree", "J")
          while 1:
            s = g.readline()
            if not s: break
            s = s.strip()
            if len(s) < 20:
              if "nstsv" in s.split():
                 nstsv = int(s.split()[0])
                 nstsv2=int(nstsv/2)
              elif "nkpt" in s.split():
                 nkpt = int(s.split()[0])
              continue
            elif len(s) > 50 and ('k-point,' in s.split()):
              try:
                int(s[0])  # assert this line is not a header string
              except ValueError:
                continue
              else:
                eigvalVal.append([])
                eigvalOcc.append([])
                eigvalKpoint.append([float(x) for x in s.split()[1:4]])

            else:
              try:
                int(s[0])
              except ValueError:
                # ignore headers by neglecting lines that
                # don't start with an integer
                continue
              else:
                # FIXME: EIG files with 'partial chg density' have 7 columns
                # FIXME: and we need to process them
                n, e, occ = s.split()[0:3]
                #---------
                try:  # eigenvalues 'e' could be wrongly formatted
                  enew = float(e)
                except Exception as ex:
                  if 'E' not in e.upper():
                    pieces = e.split('-')
                    if (len(pieces) != 2):
                      raise ex
                    try:
                        pieces = [ float(ii) for ii in pieces ]
                    except:
                        raise ex
                    mantissa, exponent = pieces
                    enew = mantissa * 10**(-1 * exponent)
                    logger.warning("In-house conversion '{}' -> {}" .format(e, enew))
                #---------
                eigvalVal[-1].append(fromH(enew))
                eigvalOcc[-1].append(float(occ))
          if not self.spinTreat:
            for i in range(0,nkpt):
              eigvalValSpin[0].append(eigvalVal[i][0:nstsv])
              eigvalOccSpin[0].append(eigvalOcc[i][0:nstsv])
              eigvalValSpin[1].append(eigvalVal[i][0:nstsv])
              eigvalOccSpin[1].append(eigvalOcc[i][0:nstsv])
            backend.addValue("eigenvalues_values", eigvalValSpin)
            backend.addValue("eigenvalues_occupation", eigvalOccSpin)
          else:
            for i in range(0,nkpt):
              eigvalValSpin[0].append(eigvalVal[i][0:nstsv2])
              eigvalOccSpin[0].append(eigvalOcc[i][0:nstsv2])
              eigvalValSpin[1].append(eigvalVal[i][nstsv2:nstsv])
              eigvalOccSpin[1].append(eigvalOcc[i][nstsv2:nstsv])
            backend.addValue("eigenvalues_values", eigvalValSpin)
            backend.addValue("eigenvalues_occupation", eigvalOccSpin)
          backend.addValue("eigenvalues_kpoints", eigvalKpoint)
          backend.closeSection("section_eigenvalues",eigvalGIndex)

    ########################## Parsing Fermi surface ##################
    if os.path.exists(fermiSurfFile):
      fermiGIndex = backend.openSection("x_exciting_section_fermi_surface")
      with open(fermiSurfFile) as g:
        grid = []
        all_vectors = []
        values = []
        origin = []
        vectors = []
        fermi = 0
        number_of_bands = 0
        mesh_size = 0
        fromH = unit_conversion.convert_unit_function("hartree", "J")
        while 1:
          s = g.readline()
          if not s: break
          s = s.strip()
          st = s.split()
          if len(st) == 3:
            if len(s) >= 40:
              all_vectors.append([])
              i = 0
              while i < 3:
                all_vectors[-1].append(float(st[i]))
                i += 1
            elif st[0] == "Fermi":
              fermi = fromH(float(st[2]))
            else:
              j = 0
              while j < 3:
                grid.append(int(st[j]))
                j += 1
          elif len(st) == 2:
            values.append([])
          elif len(s) >= 12 and len(st) == 1:
            try: float(st[0])
            except ValueError:
              continue
            else:
              values[-1].append(float(st[0]))
          elif len(s) < 5 and len(st) == 1:
            number_of_bands = st[0]
        mesh_size = grid[0]*grid[1]*grid[2]
        origin = all_vectors[0]
        vectors = all_vectors[1:]
        backend.addValue("x_exciting_number_of_bands_fermi_surface", int(number_of_bands))
        backend.addValue("x_exciting_number_of_mesh_points_fermi_surface", int(mesh_size))
        backend.addValue("x_exciting_fermi_energy_fermi_surface", float(fermi))
        backend.addArrayValues("x_exciting_grid_fermi_surface", np.asarray(grid))
        backend.addArrayValues("x_exciting_origin_fermi_surface", np.asarray(origin))
        backend.addArrayValues("x_exciting_vectors_fermi_surface", np.asarray(vectors))
        backend.addArrayValues("x_exciting_values_fermi_surface", np.asarray(values))
        backend.closeSection("x_exciting_section_fermi_surface",fermiGIndex)

  def onClose_x_exciting_section_spin(self, backend, gIndex, section):

    spin = section["x_exciting_spin_treatment"][0]
    spin = spin.strip()
    if spin == "spin-polarised":
      self.spinTreat = True
    else:
      self.spinTreat = False

  def onClose_section_system(self, backend, gIndex, section):

    self.unit_cell_vol = section["x_exciting_unit_cell_volume"]
    self.gmaxvr = section["x_exciting_gmaxvr"]
    self.rgkmax = section["x_exciting_rgkmax"]
    backend.addArrayValues('configuration_periodic_dimensions', np.asarray([True, True, True]))

    # next line fixes normalizer error "no lattice vectors but pbc"
    backend.addArrayValues("lattice_vectors", self.sim_cell)

    self.secSystemDescriptionIndex = gIndex

    if self.atom_pos and self.cell_format[0] == 'cartesian':
      #  print("self.atom_pos=",self.atom_pos)
      backend.addArrayValues('atom_positions', np.asarray(self.atom_pos))
    elif self.atom_pos and self.cell_format[0] == 'lattice':
      # print("aaaself.atom_pos=",self.atom_pos)
      # print("aaaself.atom_labels=",self.atom_labels)
       atoms = Atoms(self.atom_labels, self.atom_pos, cell=[(1, 0, 0),(0, 1, 0),(0, 0, 1)])
       atoms.set_cell(self.sim_cell, scale_atoms=True)
       self.atom_pos = atoms.get_positions()
       backend.addArrayValues('atom_positions', np.asarray(self.atom_pos))
    if self.atom_labels is not None:
      # print("aaaself.atom_labels=",self.atom_labels)
      backend.addArrayValues('atom_labels', np.asarray(self.atom_labels))
    self.atom_labels = []

    excSmearingKind = section["x_exciting_smearing_type"]

    smearing_internal_map = {
        "Gaussian": ['gaussian'],
        "Methfessel-Paxton": ['methfessel-paxton'],
        "Fermi-Dirac": ['fermi'],
        "Extended": ['tetrahedra']
        }

    if self.samplingMethod is not "geometry_optimization":
        for smName in smearing_internal_map[excSmearingKind[0]]:
          backend.addValue("smearing_kind", smName)
    else:
        pass

  def onClose_x_exciting_section_atoms_group(self, backend, gIndex, section):
    # print("start.self.atom_labels=",self.atom_labels)
    fromB = unit_conversion.convert_unit_function("bohr", "m")
    formt = section['x_exciting_atom_position_format']
    if self.samplingMethod is not "geometry_optimization":
      pass
    else:
      if self.atom_pos is not None: self.atom_pos = []
      if self.atom_labels is not None: self.atom_labels = []
    self.cell_format = formt
    pos = [section['x_exciting_geometry_atom_positions_' + i] for i in ['x', 'y', 'z']]
    # print("pos=",pos)
    pl = [len(comp) for comp in pos]
    natom = pl[0]
    if pl[1] != natom or pl[2] != natom:
      raise Exception("invalid number of atoms in various components %s" % pl)
    for i in range(natom):
      # print("nattom=",natom)
      # print("i=",i)
      # print("natom=",natom)
      # print("[pos[0][i]=",pos[0][i])
      # print("[pos[1][i]=",pos[1][i])
      # print("[pos[2][i]=",pos[2][i])
      # print("self.atom_pos=",self.atom_pos)
      if formt[0] == 'cartesian':
        self.atom_pos.append([fromB(pos[0][i]), fromB(pos[1][i]), fromB(pos[2][i])])
      else:
        # print("self.atom_labels_prima=",self.atom_labels)
        # print("self.atom_pos_prima=",self.atom_pos)
        self.atom_pos.append([pos[0][i], pos[1][i], pos[2][i]])
        # print("self.atom_pos_dopo=",self.atom_pos)
        # print("self.atom_labels_dopo=",self.atom_labels)
        # print("natom=",natom)
        # print("section['x_exciting_geometry_atom_labels']=",section['x_exciting_geometry_atom_labels'])
        # print("self.samplingMethod[0]=",self.samplingMethod)
    if self.samplingMethod is not "geometry_optimization":
        # print("prima=",self.atom_labels)
        self.atom_labels = self.atom_labels + (section['x_exciting_geometry_atom_labels'] * natom)
        # print("self.atom_labels=",self.atom_labels)
        # print("section['x_exciting_geometry_atom_labels']=",section['x_exciting_geometry_atom_labels'])
    else:
        self.atom_labels = self.atom_labels + section['x_exciting_geometry_atom_labels']
        # print("self.self.XSSetGIndexatom_labels_dopodopo=",self.atom_labels)

  def onClose_section_method(self, backend, gIndex, section):
    if gIndex == self.XSSetGIndex and self.xstype == "TDDFT":
        # tmk: if 'input.xml' is missing, then the next are 'NoneType'
        try:
          self.tddftKernel = section["x_exciting_xs_tddft_xc_kernel"]
        except:
          pass
        try:
          self.xsTetra = section["x_exciting_xs_tetra"][0]
        except:
          pass
        try:
          self.xsAC = section["x_exciting_xs_tddft_analytic_continuation"][0]
        except:
          pass
        try:
          self.xsNAR = section["x_exciting_xs_tddft_anti_resonant_dielectric"][0]
        except:
          pass
    if gIndex == self.secMethodIndex:
      backend.addValue('electronic_structure_method', "DFT")
      try:
        energy_thresh = section["x_exciting_scf_threshold_energy_change"][0]
      except:
        pass
      try:
        potential_thresh = section["x_exciting_scf_threshold_potential_change_list"][0]
      except:
        pass
      try:
        charge_thresh = section["x_exciting_scf_threshold_charge_change_list"][0]
      except:
        pass
      if section["x_exciting_scf_threshold_force_change_list"]:
        force_thresh = section["x_exciting_scf_threshold_force_change_list"][0]
        try:
          backend.addValue('x_exciting_scf_threshold_force_change', force_thresh)
        except:
          pass
      try:
        backend.addValue('scf_threshold_energy_change', energy_thresh)
      except:
        pass
      try:
        backend.addValue('x_exciting_scf_threshold_potential_change', potential_thresh)
      except:
        pass
      try:
        backend.addValue('x_exciting_scf_threshold_charge_change', charge_thresh)
      except:
        pass
      ##########BELOW VOLUME OPTIMIZATION######################
      #    if self.volumeOptIndex>1:
      #      self.volumeOpt = True
      #      optGindex = backend.openSection("section_method")
      #      backend.addValue("x_exciting_volume_optimization", self.volumeOpt)
      #      backend.closeSection("section_method", optGindex)
      #
      #    ext_uri = []
      #    backend.addValue("x_exciting_dummy", self.volumeOptIndex)
      #    if self.volumeOptIndex > 1:
      #      for j in range(1, self.volumeOptIndex):
      #        if (j<10):
      #          ext_uri.append(self.mainFilePath[0:-9] + 'rundir-0' + str(j) + '/INFO.OUT')
      #        else:
      #          ext_uri.append(self.mainFilePath[0:-9] + 'rundir-' + str(j) + '/INFO.OUT')
      #    for ref in ext_uri:
      #      refGindex = backend.openSection("section_calculation_to_calculation_refs")
      #      backend.addValue("calculation_to_calculation_external_url", ref)
      #      backend.addValue("calculation_to_calculation_kind", "source_calculation")
      #      backend.closeSection("section_calculation_to_calculation_refs", refGindex)

mainFileDescription = \
    SM(name = "root matcher",
       startReStr = "",
       weak = True,
       subMatchers = [
         SM(name = "header",
         startReStr = r"\s*(\||\+|\*)\s*EXCITING\s*(?P<program_version>[-a-zA-Z0-9]+)\s*started\s*=",
         fixedStartValues={'program_name': 'exciting', 'program_basis_set_type': '(L)APW+lo' },
            sections = ["section_run", "section_method"],
         subMatchers = [
	   SM(name = 'input',
              startReStr = r"(\||\+|\*)\sStarting initialization",
              endReStr = r"(\||\+|\*)\sEnding initialization",
              sections = ['section_system'],
              subMatchers = [
                SM(startReStr = r"\sLattice vectors \(cartesian\) :",
                sections = ["x_exciting_section_lattice_vectors"],
                subMatchers = [

    SM(startReStr = r"\s*(?P<x_exciting_geometry_lattice_vector_x__bohr>[-+0-9.]+)\s+(?P<x_exciting_geometry_lattice_vector_y__bohr>[-+0-9.]+)\s+(?P<x_exciting_geometry_lattice_vector_z__bohr>[-+0-9.]+)", repeats = True)
                ]),
                SM(startReStr = r"\sReciprocal lattice vectors \(cartesian\) :",
                sections = ["x_exciting_section_reciprocal_lattice_vectors"],
                subMatchers = [

    SM(startReStr = r"\s*(?P<x_exciting_geometry_reciprocal_lattice_vector_x__bohr_1>[-+0-9.]+)\s+(?P<x_exciting_geometry_reciprocal_lattice_vector_y__bohr_1>[-+0-9.]+)\s+(?P<x_exciting_geometry_reciprocal_lattice_vector_z__bohr_1>[-+0-9.]+)", repeats = True)
                ]),
    SM(r"\s*Unit cell volume\s*:\s*(?P<x_exciting_unit_cell_volume__bohr3>[-0-9.]+)"),
    SM(r"\s*Brillouin zone volume\s*:\s*(?P<x_exciting_brillouin_zone_volume__bohr_3>[-0-9.]+)"),
    SM(r"\s*Species\s*:\s*[0-9]\s*\((?P<x_exciting_geometry_atom_labels>[-a-zA-Z0-9]+)\)", repeats = True,
      sections = ["x_exciting_section_atoms_group"],
       subMatchers = [
        SM(r"\s*muffin-tin radius\s*:\s*(?P<x_exciting_muffin_tin_radius__bohr>[-0-9.]+)"),
        SM(r"\s*# of radial points in muffin-tin\s*:\s*(?P<x_exciting_muffin_tin_points>[-0-9.]+)"),
        SM(startReStr = r"\s*atomic positions\s*\((?P<x_exciting_atom_position_format>[-a-zA-Z]+)\)\s*:\s*",
           endReStr = r"\s*magnetic fields\s*",
           subMatchers = [
                    SM(r"\s*(?P<x_exciting_geometry_atom_number>[+0-9]+)\s*:\s*(?P<x_exciting_geometry_atom_positions_x>[-+]?[0-9.]+)\s*(?P<x_exciting_geometry_atom_positions_y>[-+]?[0-9.]+)\s*(?P<x_exciting_geometry_atom_positions_z>[-+]?[0-9.]+)", repeats = True)
         ]) #,
        # SM(startReStr = r"\s*magnetic fields\s*\((?P<x_exciting_magnetic_field_format>[-a-zA-Z]+)\)\s*:\s*",
        #   subMatchers = [
        #     SM(r"\s*(?P<x_exciting_MT_external_magnetic_field_atom_number>[+0-9]+)\s*:\s*(?P<x_exciting_MT_external_magnetic_field_x>[-+0-9.]+)\s*(?P<x_exciting_MT_external_magnetic_field_y>[-+0-9.]+)\s*(?P<x_exciting_MT_external_magnetic_field_z>[-+0-9.]+)", repeats = True)
        #  ])
    ]),
    SM(r"\s*Total number of atoms per unit cell\s*:\s*(?P<x_exciting_number_of_atoms>[-0-9.]+)"),
    SM(r"\s*Spin treatment\s*:\s*(?P<x_exciting_spin_treatment>[-a-zA-Z\s*]+)",
       sections = ["x_exciting_section_spin"]),
    SM(r"\s*k-point grid\s*:\s*(?P<x_exciting_number_kpoint_x>[-0-9.]+)\s+(?P<x_exciting_number_kpoint_y>[-0-9.]+)\s+(?P<x_exciting_number_kpoint_z>[-0-9.]+)"),
    SM(r"\s*k-point offset\s*:\s*(?P<x_exciting_kpoint_offset_x>[-0-9.]+)\s+(?P<x_exciting_kpoint_offset_y>[-0-9.]+)\s+(?P<x_exciting_kpoint_offset_z>[-0-9.]+)"),
    SM(r"\s*Total number of k-points\s*:\s*(?P<x_exciting_number_kpoints>[-0-9.]+)"),
    SM(r"\s*R\^MT_min \* \|G\+k\|_max \(rgkmax\)\s*:\s*(?P<x_exciting_rgkmax>[-0-9.]+)"),
    SM(r"\s*Maximum \|G\+k\| for APW functions\s*:\s*(?P<x_exciting_gkmax__bohr_1>[-0-9.]+)"),
    SM(r"\s*Maximum \|G\| for potential and density\s*:\s*(?P<x_exciting_gmaxvr__bohr_1>[-0-9.]+)"),
    SM(r"\s*G-vector grid sizes\s*:\s*(?P<x_exciting_gvector_size_x>[-0-9.]+)\s+(?P<x_exciting_gvector_size_y>[-0-9.]+)\s+(?P<x_exciting_gvector_size_z>[-0-9.]+)"),
    SM(r"\s*Total number of G-vectors\s*:\s*(?P<x_exciting_gvector_total>[-0-9.]+)"),
    SM(startReStr = r"\s*Maximum angular momentum used for\s*",
        subMatchers = [
          SM(r"\s*APW functions\s*:\s*(?P<x_exciting_lmaxapw>[-0-9.]+)")
        ]),
    SM(r"\s*Total nuclear charge\s*:\s*(?P<x_exciting_nuclear_charge>[-0-9.]+)"),
    SM(r"\s*Total electronic charge\s*:\s*(?P<x_exciting_electronic_charge>[-0-9.]+)"),
    SM(r"\s*Total core charge\s*:\s*(?P<x_exciting_core_charge_initial>[-0-9.]+)"),
    SM(r"\s*Total valence charge\s*:\s*(?P<x_exciting_valence_charge_initial>[-0-9.]+)"),
    SM(r"\s*Effective Wigner radius, r_s\s*:\s*(?P<x_exciting_wigner_radius>[-0-9.]+)"),
    SM(r"\s*Number of empty states\s*:\s*(?P<x_exciting_empty_states>[-0-9.]+)"),
    SM(r"\s*Total number of valence states\s*:\s*(?P<x_exciting_valence_states>[-0-9.]+)"),
    SM(r"\s*Maximum Hamiltonian size\s*:\s*(?P<x_exciting_hamiltonian_size>[-0-9.]+)"),
    SM(r"\s*Maximum number of plane-waves\s*:\s*(?P<x_exciting_pw>[-0-9.]+)"),
    SM(r"\s*Total number of local-orbitals\s*:\s*(?P<x_exciting_lo>[-0-9.]+)"),
    SM(startReStr = r"\s*Exchange-correlation type\s*:\s*(?P<x_exciting_xc_functional>[-0-9.]+)",
       sections = ['x_exciting_section_xc']),
    SM(r"\s*Smearing scheme\s*:\s*(?P<x_exciting_smearing_type>[-a-zA-Z0-9]+)"),
    SM(r"\s*Smearing width\s*:\s*(?P<smearing_width__hartree>[-0-9.]+)"),
    SM(r"\s*Using\s*(?P<x_exciting_potential_mixing>[-a-zA-Z\s*]+)\s*potential mixing")
    ]),
            SM(name = "single configuration iteration",
              startReStr = r"(\||\+|\*)\s*Self-consistent loop started\s*\+",
              sections = ["section_single_configuration_calculation"],
              repeats = True,
              subMatchers = [
                SM(name = "scfi totE",
                 startReStr =r"(\||\+|\*)\s*SCF iteration number\s*:",
                  sections = ["section_scf_iteration"],
                  repeats = True,
                  subMatchers = [
                   SM(r"\s*Total energy\s*:\s*(?P<energy_total_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Fermi energy\s*:\s*(?P<x_exciting_fermi_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Kinetic energy\s*:\s*(?P<electronic_kinetic_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Coulomb energy\s*:\s*(?P<x_exciting_coulomb_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Exchange energy\s*:\s*(?P<x_exciting_exchange_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Correlation energy\s*:\s*(?P<x_exciting_correlation_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Sum of eigenvalues\s*:\s*(?P<energy_sum_eigenvalues_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Effective potential energy\s*:\s*(?P<x_exciting_effective_potential_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Coulomb potential energy\s*:\s*(?P<x_exciting_coulomb_potential_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*xc potential energy\s*:\s*(?P<energy_XC_potential_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Hartree energy\s*:\s*(?P<x_exciting_hartree_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Electron-nuclear energy\s*:\s*(?P<x_exciting_electron_nuclear_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Nuclear-nuclear energy\s*:\s*(?P<x_exciting_nuclear_nuclear_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Madelung energy\s*:\s*(?P<x_exciting_madelung_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Core-electron kinetic energy\s*:\s*(?P<x_exciting_core_electron_kinetic_energy_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Absolute change in total energy   (target)\s*:\s*(?P<energy_change_scf_iteration__hartree>[-0-9.]+)\s*(\s*(?P<scf_threshold_energy_change__hartree>[-0-9.]+))"),
                   SM(r"\s*DOS at Fermi energy \(states\/Ha\/cell\)\s*:\s*(?P<x_exciting_dos_fermi_scf_iteration__hartree_1>[-0-9.]+)"),
                   SM(r"\s*core\s*:\s*(?P<x_exciting_core_charge_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*core leakage\s*:\s*(?P<x_exciting_core_leakage_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*valence\s*:\s*(?P<x_exciting_valence_charge_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*interstitial\s*:\s*(?P<x_exciting_interstitial_charge_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*total charge in muffin-tins\s*:\s*(?P<x_exciting_total_MT_charge_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*Estimated fundamental gap\s*:\s*(?P<x_exciting_gap_scf_iteration__hartree>[-0-9.]+)"),
                   SM(r"\s*Wall time \(seconds\)\s*:\s*(?P<x_exciting_time_scf_iteration>[-0-9.]+)"),
                   SM(r"\s*RMS change in effective potential \(target\)\s*:\s*(?P<x_exciting_effective_potential_convergence_scf_iteration__hartree>[0-9]+\.[0-9]*([E]?[-]?[0-9]+))\s*\(\s*(?P<x_exciting_scf_threshold_potential_change_list__hartree>[0-9]\.[0-9]*([E]?[-]?[0-9]+))\)"),
                   SM(r"\s*Absolute change in total energy\s*\(target\)\s*:\s*(?P<x_exciting_energy_convergence_scf_iteration>[0-9]+\.[0-9]*([E]?[-]?[0-9]+))\s*\(\s*(?P<x_exciting_scf_threshold_energy_change__hartree>[0-9]\.[0-9]*([E]?[-]?[0-9]+))\)"),
                   SM(r"\s*Absolute change in total energy\s*\(target\)\s*\:\s*\s*\(\s*[0-9]+\.[0-9]+\s*[0-9]+\.[0-9]+\s*[0-9]+\.[0-9]+\s*\)\s*avg =\s*[0-9]+\.[0-9]+\s*\(\s*(?P<x_exciting_scf_threshold_energy_change__hartree>[0-9.]+E?[-][0-9]+)\s*\)"),
                   SM(r"\s*Charge distance\s*\(target\)\s*:\s*(?P<x_exciting_charge_convergence_scf_iteration>[0-9]\.[0-9]*([E]?[-]?[0-9]+))\s*\(\s*(?P<x_exciting_scf_threshold_charge_change_list>[0-9]\.[0-9]*([E]?[-]?[0-9]+))\)"),
                   SM(r"\s*Abs. change in max-nonIBS-force\s*\(target\)\s*:\s*(?P<x_exciting_force_convergence_scf_iteration>[0-9]\.[0-9]*([E]?[-]?[0-9]+))\s*\(\s*(?P<x_exciting_scf_threshold_force_change_list>[0-9]\.[0-9]*([E]?[-]?[0-9]+))\)")
                  ]),
                SM(name="final_quantities",
                  startReStr = r"(\||\+|\*) Convergence targets achieved. Performing final SCF iteration\s*\+",
                  endReStr = r"(\||\+|\*) Self-consistent loop stopped\s*\+",
                   subMatchers = [
                     SM(r"\s*Total energy\s*:\s*(?P<energy_total__hartree>[-0-9.]+)"),
                     SM(r"\s*Fermi energy\s*:\s*(?P<x_exciting_fermi_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Kinetic energy\s*:\s*(?P<electronic_kinetic_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Coulomb energy\s*:\s*(?P<x_exciting_coulomb_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Exchange energy\s*:\s*(?P<x_exciting_exchange_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Correlation energy\s*:\s*(?P<x_exciting_correlation_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Sum of eigenvalues\s*:\s*(?P<energy_sum_eigenvalues__hartree>[-0-9.]+)"),
                     SM(r"\s*Effective potential energy\s*:\s*(?P<x_exciting_effective_potential_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Coulomb potential energy\s*:\s*(?P<x_exciting_coulomb_potential_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*xc potential energy\s*:\s*(?P<energy_XC_potential__hartree>[-0-9.]+)"),
                     SM(r"\s*Hartree energy\s*:\s*(?P<x_exciting_hartree_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Electron-nuclear energy\s*:\s*(?P<x_exciting_electron_nuclear_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Nuclear-nuclear energy\s*:\s*(?P<x_exciting_nuclear_nuclear_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Madelung energy\s*:\s*(?P<x_exciting_madelung_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*Core-electron kinetic energy\s*:\s*(?P<x_exciting_core_electron_kinetic_energy__hartree>[-0-9.]+)"),
                     SM(r"\s*DOS at Fermi energy \(states\/Ha\/cell\)\s*:\s*(?P<x_exciting_dos_fermi__hartree_1>[-0-9.]+)"),
                     SM(r"\s*core leakage\s*:\s*(?P<x_exciting_core_leakage>[-0-9.]+)"),
                     SM(r"\s*interstitial\s*:\s*(?P<x_exciting_interstitial_charge>[-0-9.]+)"),
                     SM(r"\s*total charge in muffin-tins\s*:\s*(?P<x_exciting_total_MT_charge>[-0-9.]+)"),
                     SM(r"\s*Estimated fundamental gap\s*:\s*(?P<x_exciting_gap__hartree>[-0-9.]+)")
                   ]) #,
                #                SM(name="final_forces",
                ##                  startReStr = r"(\||\+|\*) Writing atomic positions and forces\s*\-",
                #                  startReStr = r"\s*Total atomic forces including IBS \(cartesian\) \s*:",
                #                  endReStr = r"(\||\+|\*)\s*Groundstate module stopped\s*\*",
                ##                  endReStr = r"\s* Atomic force components including IBS \(cartesian\)\s*:",
                #                  floating = True,
                #                   subMatchers = [
                ##                     SM(name="total_forces",
                ##                     startReStr = r"\s*Total atomic forces including IBS \(cartesian\)\s*:",
                #                       SM(r"\s*atom\s*[0-9]+\s*[A-Za-z]+\s*\:\s*(?P<x_exciting_atom_forces_x>[-0-9.]+)\s*(?P<x_exciting_atom_forces_y>[-0-9.]+)\s*(?P<x_exciting_atom_forces_z>[-0-9.]+)",
                #                          repeats = True )
                ######                     subMatchers = [
                ######                     SM(r"\s*atom\s*(?P<x_exciting_store_total_forces>[0-9]+\s*[A-Za-z]+\s*\:+\s*[-\d\.]+\s*[-\d\.]+\s*[-\d\.]+)",
                ######                          repeats = True)
                ######                   ] )
                ##)
                ##                     print ("number atoms=", x_exciting_number_of_atoms)
                ##                     SM(name="force_components",
                ##                     startReStr = r"\s*Atomic force components including IBS \(cartesian\)\s*:",
                ##                     forwardMatch = True,
                ##                     subMatchers = [
                ##                     SM(r"\s*atom\s*(?P<x_exciting_store_total_forces>[0-9]+\s*[A-Za-z]+\s*\:+\s*[-\d\.]+\s*[-\d\.]+\s*[-\d\.]+\s*[A-Za-z]+\s*[A-Za-z]+)", weak = True),
                ##                     SM(r"\s*(?P<x_exciting_store_total_forces>\s*\:+\s*[-\d\.]+\s*[-\d\.]+\s*[-\d\.]+\s*[A-Za-z]+\s*[A-Za-z]+)"),
                ##                     SM(r"\s*(?P<x_exciting_store_total_forces>\s*\:+\s*[-\d\.]+\s*[-\d\.]+\s*[-\d\.]+\s*[A-Za-z]+\s*[A-Za-z]+)")
                ##                     SM(r"\s*(?P<x_exciting_store_total_forces>\s*\:+\s*[-\d\.]+\s*[-\d\.]+\s*[-\d\.]+\s*[A-Za-z]+\s*[A-Za-z]+)"),
                ##                   ]
                ##                    )
                #                   ]),
                #                 SM(name="force_components",
                #                  startReStr = r"\s* Atomic force components including IBS \(cartesian\)\s*:",
                #                  endReStr = r"(\||\+|\*)\s* Groundstate module stopped\s* \*",
                #                  subMatchers = [
                ##                  startReStr = r"\s* Atomic force components including IBS \(cartesian\)\s*:",
                #                   SM(r"\s*atom\s*[0-9]+\s*[A-Za-z]+\s*\:\s*(?P<x_exciting_atom_HF_forces_x>[-0-9.]+)\s*(?P<x_exciting_atom_HF_forces_y>[-0-9.]+)\s*(?P<x_exciting_atom_HF_forces_z>[-0-9.]+)\s*HF force",
                #                     repeats = True,
                #                     floating = True),
                #                   SM(r"\s*\:\s*(?P<x_exciting_atom_core_forces_x>[-0-9.]+)\s*(?P<x_exciting_atom_core_forces_y>[-0-9.]+)\s*(?P<x_exciting_atom_core_forces_z>[-0-9.]+)\s*core correction",
                #                     repeats = True,
                #                     floating = True),
                #                   SM(r"\s*\:\s*(?P<x_exciting_atom_IBS_forces_x>[-0-9.]+)\s*(?P<x_exciting_atom_IBS_forces_y>[-0-9.]+)\s*(?P<x_exciting_atom_IBS_forces_z>[-0-9.]+)\s*IBS correction",
                #                     repeats = True,
                #                     floating = True),
                ##                   SM(r"(?P<x_exciting_store_total_forces>.*)",
                ##                          repeats = True,
                #                ] )
               ]
            ),
            SM(name = "geometry optimization",
              startReStr = r"(\||\+|\*)\s*Structure-optimization module started*\s*\*",
              sections = ["section_sampling_method","x_exciting_section_geometry_optimization"],
              # fixedStartValues={'sampling_method': 'geometry_optimization'},
              # repeats = True,
              subMatchers = [
                   SM(name = "optimization steps",
                   startReStr = r"(\||\+|\*)\s*Optimization step\s*(?P<x_exciting_geometry_optimization_step>[-0-9]+)\s*\(method = (?P<x_exciting_geometry_optimization_method>[A-Za-z]+)\)\s*\-",
                   sections = ["section_single_configuration_calculation"],
                  # SM(r"\s*Output level for this task is set to normal\s*"),
                  # SM(r"(\||\+|\*)\s*Optimization step (?P<x_exciting_geometry_optimization_step>[-0-9]+)\: Initialize optimization\s*\-"),
                   repeats = True,
                   subMatchers = [
                   SM(r"\s*Maximum force magnitude\s*\(target\)\s*:\s*(?P<x_exciting_maximum_force_magnitude__hartree_bohr_1>[0-9]+\.[0-9]*([E]?[-]?[0-9]+))\s*\(\s*(?P<x_exciting_geometry_optimization_threshold_force__hartree_bohr_1>[0-9]\.[0-9]*([E]?[-]?[0-9]+))\)"),
                   SM(r"\s*Total energy at this optimization step\s*:\s*(?P<energy_total__hartree>[-0-9.]+)"),
                   SM(startReStr = r"\s*Atomic positions at this step \s*\((?P<x_exciting_atom_position_format>[-a-zA-Z]+)\)\s*:\s*",
                  # endReStr = r"\s*Total atomic forces including IBS \(cartesian\) \:",
                  # weak = True,
                   sections = ["section_system","x_exciting_section_atoms_group"],
                  # endReStr = r"\s*magnetic fields\s*",
           subMatchers = [
                    SM(r"\s*atom\s*(?P<x_exciting_atom_number>[+0-9]+)\s*(?P<x_exciting_geometry_atom_labels>[A-Za-z]+)\s*\:\s*(?P<x_exciting_geometry_atom_positions_x>[-+0-9.]+)\s*(?P<x_exciting_geometry_atom_positions_y>[-+0-9.]+)\s*(?P<x_exciting_geometry_atom_positions_z>[-+0-9.]+)", repeats = True)
         ]),
                   SM(startReStr = r"\s*Total atomic forces including IBS \(cartesian\) \:",
                    endReStr = r"\s*Time spent in this optimization step\s*\:\s*(?P<x_exciting_geometry_dummy>[+0-9.]+)\s*seconds",
                    subMatchers = [
                    SM(r"\s*atom\s*[+0-9]+\s*[A-Za-z]+\s*\:\s*(?P<x_exciting_geometry_atom_forces_x__hartree_bohr_1>[-+0-9.]+)\s*(?P<x_exciting_geometry_atom_forces_y__hartree_bohr_1>[-+0-9.]+)\s*(?P<x_exciting_geometry_atom_forces_z__hartree_bohr_1>[-+0-9.]+)", repeats = True)
         ])
         ])
           ]
           )
          ])
    ])




parserInfo = {
  "name": "exciting_parser",
  "version": "1.0"
}

cachingLevelForMetaName = {
    "x_exciting_geometry_lattice_vector_x":CachingLevel.Cache,
    "x_exciting_geometry_lattice_vector_y":CachingLevel.Cache,
    "x_exciting_geometry_lattice_vector_z":CachingLevel.Cache,
    "x_exciting_section_lattice_vectors": CachingLevel.Ignore,
    "x_exciting_geometry_reciprocal_lattice_vector_x":CachingLevel.Cache,
    "x_exciting_geometry_reciprocal_lattice_vector_y":CachingLevel.Cache,
    "x_exciting_geometry_reciprocal_lattice_vector_z":CachingLevel.Cache,
    "x_exciting_section_reciprocal_lattice_vectors": CachingLevel.Ignore,
    "x_exciting_atom_forces_x":CachingLevel.Cache,
    "x_exciting_atom_forces_y":CachingLevel.Cache,
    "x_exciting_atom_forces_z":CachingLevel.Cache,
    "x_exciting_atom_HF_forces_x":CachingLevel.Cache,
    "x_exciting_atom_HF_forces_y":CachingLevel.Cache,
    "x_exciting_atom_HF_forces_z":CachingLevel.Cache,
    "x_exciting_atom_core_forces_x":CachingLevel.Cache,
    "x_exciting_atom_core_forces_y":CachingLevel.Cache,
    "x_exciting_atom_core_forces_z":CachingLevel.Cache,
    "x_exciting_atom_IBS_forces_x":CachingLevel.Cache,
    "x_exciting_atom_IBS_forces_y":CachingLevel.Cache,
    "x_exciting_atom_IBS_forces_z":CachingLevel.Cache,
    "x_exciting_geometry_atom_forces_x":CachingLevel.Cache,
    "x_exciting_geometry_atom_forces_y":CachingLevel.Cache,
    "x_exciting_geometry_atom_forces_z":CachingLevel.Cache
 }


class ExcitingParser():
    """ A proper class envolop for running this parser from within python. """
    def __init__(self, backend, **kwargs):
        self.backend_factory = backend

    def parse(self, mainfile):
        from unittest.mock import patch
        logging.info('exciting parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend_factory('exciting.nomadmetainfo.json')
        with patch.object(sys, 'argv', ['<exe>', '--uri', 'nmd://uri', mainfile]):
            mainFunction(
                mainFileDescription,
                None,
                parserInfo,
                cachingLevelForMetaName = cachingLevelForMetaName,
                superContext=ExcitingParserContext(),
                superBackend=backend)
            # print("#"*50 + "\nPARSING ENDED. NORMALIZER FOLLOWS..."+"\n"*5)# tmk:
        return backend



import calendar
import json
import os
import sys
import re
import numpy as np
import logging
from nomadcore.local_meta_info import loadJsonFile, InfoKindEl
from nomadcore.unit_conversion.unit_conversion import convert_unit
from nomadcore.simple_parser import mainFunction, SimpleMatcher as SM, CachingLevel
from nomadcore.baseclasses import ParserInterface

############################################################
# This file contains functions and constants that are needed
# by more than one parser.
############################################################


LOGGER = logging.getLogger(__name__)

# fortran float, alternate too-long-for-field fortran marker
RE_f = r"(?:[+-]?\d+(?:\.\d+)?(?:[eEdD][+-]?\d+)?|\*+)"
cRE_f = re.compile(RE_f)
# fortran int, alternate too-long-for-field fortran marker
RE_i = r"(?:[+-]?\d+|\*+)"
cRE_i = re.compile(RE_i)
NAN = float('nan')

def re_vec(name, units='', split="\s+"):
    """generator for 3-component vector regex"""
    if units:
        units = '__' + units
    res = (
        r'(?P<' + name + r'_x' + units + r'>' + RE_f + r')' + split +
        r'(?P<' + name + r'_y' + units + r'>' + RE_f + r')' + split +
        r'(?P<' + name + r'_z' + units + r'>' + RE_f + r')'
        )
    return res


PARSER_INFO_DEFAULT = {
  "name": "parser_quantum_espresso",
  "version": "0.0.1"
}

# constants for date conversion
MONTHS = [ 'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
           'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec' ]
MONTH_NUMBER = { MONTHS[num]: num+1 for num in range(0,12) }


QE_SMEARING_KIND = {
    '-99': 'fermi',
     '-1': 'marzari-vanderbilt',
      '0': 'gaussian',
      '1': 'methfessel-paxton',
    'Marzari-Vanderbilt smearing': 'marzari-vanderbilt',
     'Methfessel-Paxton smearing': 'methfessel-paxton',
              'gaussian smearing': 'gaussian',
           'Fermi-Dirac smearing': 'fermi',
             'tetrahedron method': 'tetrahedra',
}


class ParserQuantumEspresso():
    """Base class for all Quantum Espresso parsers"""
    def __init__(
        self, cachingLevelForMetaName=None, coverageIgnoreList=None,
        re_program_name=None, metainfo_to_keep=None, backend=None, default_units=None,
        metainfo_units=None, debug=True, log_level=logging.ERROR, store=True):

        self.re_program_name = re_program_name
        self.parserInfo = PARSER_INFO_DEFAULT.copy()
        self.cachingLevelForMetaName = {}
        self.backend = backend

        # common prosa in espresso output
        self.coverageIgnoreList = [
            # ignore empty lines
            r"\s*",
            # table separators
            r"^\s*[=%-]+\s*$",
            r"^\s*%\s*%\s*$",
        ]
        self.coverageIgnore = None

    def parse(self, mainfile):
        self.coverageIgnore = re.compile(r"^(?:" + r"|".join(self.coverageIgnoreList) + r")$")
        logging.info('quantum espresso parser started')
        logging.getLogger('nomadcore').setLevel(logging.WARNING)
        backend = self.backend('quantum_espresso.nomadmetainfo.json')

        for name in backend.metaInfoEnv().infoKinds:
            # set all temporaries to caching-only
            if name.startswith('x_qe_t_'):
                self.cachingLevelForMetaName[name] = CachingLevel.Cache

        mainFunction(
            self.mainFileDescription(),
            None,
            self.parserInfo,
            cachingLevelForMetaName=self.cachingLevelForMetaName,
            superContext=self,
            superBackend=backend,
            mainFile=mainfile)
        return backend

    def adHoc_suicide_qe_program_name(self, parser):
        if self.re_program_name is not None:
            if not self.re_program_name.match(
                    parser.lastMatch['x_qe_program_name']):
                raise Exception(
                    "mainFile program name was: %s, unsuited for %s" % (
                        parser.lastMatch['x_qe_program_name'],
                        type(self).__name__))

    def mainFileDescription(self):
        # assemble matchers and submatchers
        result = SM(
            name='root',
            weak=True,
            startReStr="",
            subMatchers=[
                SM(name='ktab_cIgn', coverageIgnore=True,
                   # early output seen in benchmark.out.v5.3.0.inp\=vdw1.in.1452257026
                   startReStr=r"\s*Generating kernel table - May take several minutes.*$",
                ),
                SM(name='newRun', repeats=True, required=True,
                   startReStr=(
                       # program name, e.g. PWSCF, DOS
                       r"\s*Program\s+(?P<x_qe_program_name>\S+)\s+v\." +
                       # version
                       r"(?P<program_version>\S+(?:\s+\(svn\s+rev\.\s+\d*(:\d*)?\s*\))?)" +
                       r"\s+starts" +
                       # newer espresso: "on $date"
                       r"(?:(?:\s+on\s+(?P<time_run_date_start__strQeDate>.+?)?)\s*$|" +
                       # older espresso has just "..." and date on new line
                       r"(?:\s*\.\.\.)\s*$)"
                   ),
                   adHoc = self.adHoc_suicide_qe_program_name,
                   fixedStartValues={'program_name': 'Quantum Espresso',
                                     'program_basis_set_type': 'plane waves'},
                   sections=['section_run'],
                   subMatchers=([
                       # older espresso versions have start date on separate line
                       SM(name='run_date',
                          startReStr=r"\s*Today is\s*(?P<time_run_date_start__strQeDate>.+?)\s*$"
                       ),
                       SM(name='copyright_msg', coverageIgnore=True,
                          # ignore copyright/citation msg
                          startReStr=r"\s*This program is part of the open-source Quantum ESPRESSO suite",
                          subMatchers=[
                              SM(name='copyright_msg010', coverageIgnore=True,
                                 startReStr=r"\s*for quantum simulation of materials; please (?:cite|acknowledge)",
                              ),
                              SM(name='copyright_msg020', coverageIgnore=True,
                                 startReStr=r"\s*\"P. Giannozzi et al., J. Phys.:Condens. Matter 21 395502 \(2009\);\s*",
                              ),
                              SM(name='copyright_msg030', coverageIgnore=True,
                                 startReStr=r"\s*URL http://www.quantum-espresso.org\",\s*",
                              ),
                              SM(name='copyright_msg040', coverageIgnore=True,
                                 startReStr=r"\s*in publications or presentations arising from this work. More details at",
                              ),
                              # Changed CI to False
                              SM(name='copyright_msg050', coverageIgnore=True,
                                 startReStr=r"\s*http://www.quantum-espresso.org/quote(?:\.php)?",
                              ),
                              SM(name='copyright_msg055', coverageIgnore=True,
                                 startReStr=r"\s*http://www.quantum-espresso.org/wiki/index.php/Citing_Quantum-ESPRESSO\s*",
                              ),
                          ],
                       ),
                   ] + self.run_submatchers() + [
                       SM(name='end_date',
                          startReStr=r"\s*This run was terminated on:\s*(?P<time_run_date_end__strQeDate>.+?)\s*$",
                       ),
                       SM(name='job_done',
                          startReStr=r"\s*JOB DONE\.\s*",
                          adHoc=lambda p: p.backend.addValue('run_clean_end', True),
                       ),
                   ]),
                )
            ]
        )
        return result

    def run_submatchers(self):
        return []

    def strValueTransform_strQeDate(self, espresso_date):
        if espresso_date is None:
            return None
        epoch = 0
        match = re.match(
            r"(\d+)\s*([A-Za-z]+)\s*(\d+)\s+at\s+(\d+):\s*(\d+):\s*(\d+)",
            espresso_date)
        if match:
            month = MONTH_NUMBER[match.group(2)]
            epoch = calendar.timegm(
                (int(match.group(3)), int(month), int(match.group(1)),
                 int(match.group(4)), int(match.group(5)), int(match.group(6))))
        else:
            match = re.match(
                r"\s*(\d+):\s*(\d+):\s*(\d+)\s+(\d+)\s*([A-Za-z]+)\s*(\d+)\s*",
                espresso_date)
            if match:
                month = MONTH_NUMBER[match.group(5)]
                epoch = calendar.timegm(
                    (int(match.group(6)), int(month), int(match.group(4)),
                     int(match.group(1)), int(match.group(2)), int(match.group(3))))
            else:
                raise RuntimeError("unparsable date: %s", espresso_date)
        return(epoch)
    strValueTransform_strQeDate.units = 's'

    def strValueTransform_strQeTimespan(self, espresso_timespan):
        if espresso_timespan is None:
            return None
        match = ParserQuantumEspresso.strValueTransform_strQeTimespan.re.match(
            espresso_timespan)
        if not match:
            raise RuntimeError(
                "unparsable timespan (regex match failed): %s",
                espresso_timespan)
        had_time_components = 0
        timespan_seconds = 0.0
        if match.group('seconds') is not None:
            timespan_seconds += float(match.group('seconds'))
            had_time_components += 1
        if match.group('minutes') is not None:
            timespan_seconds += float(match.group('minutes'))*60
            had_time_components += 1
        if match.group('hours') is not None:
            timespan_seconds += float(match.group('hours'))*60*60
            had_time_components += 1
        if match.group('days') is not None:
            timespan_seconds += float(match.group('days'))*60*60*24
            had_time_components += 1
        if had_time_components == 0:
            raise RuntimeError(
                "unparsable timespan (no time components extracted): %s",
                espresso_timespan)
        return(timespan_seconds)
    strValueTransform_strQeTimespan.re = re.compile(
            r"(?:\s*(?P<days>\d+)\s*d)?" +
            r"(?:\s*(?P<hours>\d+)\s*h)?" +
            r"(?:\s*(?P<minutes>\d+)\s*m)?" +
            r"(?:\s*(?P<seconds>" + RE_f + ")\s*s)?" +
            r"\s*$" # all groups are optional, end-anchor with rubber-space
    )
    strValueTransform_strQeTimespan.units = 's'

    def addDict(self, backend, this_dict):
        for key, value in sorted(this_dict.items()):
            backend.addValue(key, value)

    def addSectionDict(self, backend, section_name, section_dict):
        gIndex = backend.openSection(section_name)
        self.addDict(backend, section_dict)
        backend.closeSection(section_name, gIndex)


from builtins import object
import json
from nomadcore.compact_sha import sha512
import numpy

"""Various functions to simplify and standardize dumping objects to json"""

def numpyEncoder(o):
    """new default function for json class so that numpy arrays and sets can be encoded"""
    # check if object is a numpy array
    if isinstance(o, numpy.ndarray):
        # ensure that we have an array with row-major memory order (C like)
        if not o.flags['C_CONTIGUOUS']:
            o = numpy.ascontiguousarray(o)
        return o.tolist()
        # see default method in python/json/encoder.py
    elif isinstance(o, set):
        return list(sorted(o))
    else:
        raise TypeError(repr(o) + " is not JSON serializable")

class ExtraIndenter(object):
    """Helper class to add extra indent at the beginning of every line"""
    def __init__(self, fStream, extraIndent):
        self.fStream = fStream
        self.indent = " " * extraIndent if extraIndent else  ""
    def write(self, val):
        i = 0
        while True:
            j = val.find("\n", i)
            if j == -1:
                self.fStream.write(val[i:])
                return
            j += 1
            self.fStream.write(val[i:j])
            self.fStream.write(self.indent)
            i = j

def jsonCompactF(obj, fOut, check_circular = False):
    """Dumps the object obj with a compact json representation using the utf_8 encoding 
    to the file stream fOut"""
    json.dump(obj, fOut, sort_keys = True, indent = None, separators = (',', ':'),
            ensure_ascii = False, check_circular = check_circular, default = numpyEncoder)

def jsonIndentF(obj, fOut, check_circular = False, extraIndent = None):
    """Dumps the object obj with an indented json representation using the utf_8 encoding 
    to the file stream fOut"""
    fStream = fOut
    if extraIndent:
        fStream = ExtraIndenter(fOut, extraIndent = extraIndent)
    json.dump(obj, fStream, sort_keys = True, indent = 2, separators = (',', ': '),
            ensure_ascii = False, check_circular = check_circular, default = numpyEncoder)

class DumpToStream(object):
    """transform a dump function in a stream"""
    def __init__(self, dumpF, extraIndent = None):
        self.baseDumpF = dumpF
        self.extraIndent = extraIndent
        self.indent = " " * extraIndent if extraIndent else  ""
        self.dumpF = self.dumpIndented if extraIndent else dumpF
    def dumpIndented(self, val):
        if type(val) == type(u""):
            val = val.encode("utf_8")
        i = 0
        while True:
            j = val.find("\n", i)
            if j == -1:
                self.baseDumpF(val[i:])
                return
            j += 1
            self.baseDumpF(val[i:j])
            self.baseDumpF(self.indent)
            i = j
    def write(self, val):
        self.dumpF(val)

def jsonCompactD(obj, dumpF, check_circular = False):
    """Dumps the object obj with a compact json representation using the utf_8 encoding 
    to the file stream fOut"""
    json.dump(obj, DumpToStream(dumpF), sort_keys = True, indent = None, separators = (', ', ': '),
            ensure_ascii = False, check_circular = check_circular, default = numpyEncoder)

def jsonIndentD(obj, dumpF, check_circular = False, extraIndent = None):
    """Dumps the object obj with an indented json representation using the utf_8 encoding 
    to the function dumpF"""
    json.dump(obj, DumpToStream(dumpF, extraIndent = extraIndent), sort_keys = True, indent = 2, separators = (',', ': '),
            ensure_ascii = False, check_circular = check_circular, encoding="utf_8", default = numpyEncoder)

def jsonCompactS(obj, check_circular = False):
    """returns a compact json representation of the object obj as a string"""
    return json.dumps(obj, sort_keys = True, indent = None, separators = (', ', ': '),
            ensure_ascii = False, check_circular = check_circular, encoding="utf_8", default = numpyEncoder)

def jsonIndentS(obj, check_circular = False, extraIndent = None):
    """retuns an indented json representation if the object obj as a string"""
    res = json.dumps(obj, sort_keys = True, indent = 2, separators = (',', ': '),
            ensure_ascii = False, check_circular = check_circular, encoding="utf_8", default = numpyEncoder)
    if extraIndent:
        indent = " " * extraIndent
        res = res.replace("\n", "\n" + indent)
    return res

def jsonDump(obj, path):
    """Dumps the object obj to an newly created utf_8 file at path"""
    kwds = dict()
    if sys.version_info.major > 2:
        kwds["encoding"] = "utf_8"
    with open(path, "w", **kwds) as f:
        jsonIndentF(obj, f)

class ShaStreamer(object):
    """a file like object that calculates one or more shas"""
    def __init__(self, shas = None):
        self.shas = shas
        if shas is None:
            self.shas = (sha512(),)
    def write(self, val):
        for sha in self.shas:
            sha.update(val)
    def b64digests(self):
        return [sha.b64digest() for sha in self.shas]

def addShasOfJson(obj, shas = None):
    """adds the jsonDump of obj to the shas"""
    streamer = ShaStreamer(shas)
    jsonCompactF(obj, streamer)
    return streamer

def normalizedJsonGid(obj, shas = None):
    """returns the gid of the standard formatted jsonDump of obj"""
    return ['j' + x for x in addShasOfJson(shas).b64digests()]



# pylint: disable=line-too-long,no-self-argument
from datetime import datetime
from typing import Optional, Dict, List
from pydantic import BaseModel, Field, validator  # pylint: disable=no-name-in-module

from .jsonapi import Relationships, Attributes, Resource
from .optimade_json import Relationship


__all__ = (
    "EntryRelationships",
    "EntryResourceAttributes",
    "EntryResource",
    "EntryInfoProperty",
    "EntryInfoResource",
)


class TypedRelationship(Relationship):
    # This may be updated when moving to Python 3.8
    @validator("data")
    def check_rel_type(cls, data):
        if hasattr(cls, "_req_type") and any(obj.type != cls._req_type for obj in data):
            raise ValueError("Object stored in relationship data has wrong type")
        return data


class ReferenceRelationship(TypedRelationship):
    _req_type = "references"


class StructureRelationship(TypedRelationship):
    _req_type = "structures"


class EntryRelationships(Relationships):
    """This model wraps the JSON API Relationships to include type-specific top level keys. """

    references: Optional[ReferenceRelationship] = Field(
        None,
        description="Object containing links to relationships with entries of the `references` type.",
    )

    structures: Optional[StructureRelationship] = Field(
        None,
        description="Object containing links to relationships with entries of the `structures` type.",
    )


class EntryResourceAttributes(Attributes):
    """Contains key-value pairs representing the entry's properties."""

    immutable_id: Optional[str] = Field(
        None,
        description="""The entry's immutable ID (e.g., an UUID).
This is important for databases having preferred IDs that point to "the latest version" of a record, but still offer access to older variants.
This ID maps to the version-specific record, in case it changes in the future.
- **Type**: string.
- **Requirements/Conventions**:

  - **Support**: OPTIONAL, i.e., MAY be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter features.

- **Examples**:

  - :val:`"8bd3e750-b477-41a0-9b11-3a799f21b44f"`
  - :val:`"fjeiwoj,54;@=%<>#32"` (Strings that are not URL-safe are allowed.)""",
    )

    last_modified: datetime = Field(
        ...,
        description="""Date and time representing when the entry was last modified.
- **Type**: timestamp.
- **Requirements/Conventions**:

  - **Support**: SHOULD be supported, i.e., SHOULD NOT be :val:`null`. Is REQUIRED in this implementation, i.e., MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter features.
  - **Response**: REQUIRED in the response unless the query parameter :query-param:`response_fields` is present and does not include this property.

- **Example**:

  - As part of JSON response format: :VAL:`"2007-04-05T14:30Z"` (i.e., encoded as an `RFC 3339 Internet Date/Time Format <https://tools.ietf.org/html/rfc3339#section-5.6>`__ string.)""",
    )


class EntryResource(Resource):

    id: str = Field(
        ...,
        description="""An entry's ID as defined in section `Definition of Terms`_.
- **Type**: string.
- **Requirements/Conventions**:

  - **Support**: REQUIRED, MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter features.
  - **Response**: REQUIRED in the response.
  - See section `Definition of Terms`_.

- **Examples**:

  - :val:`"db/1234567"`
  - :val:`"cod/2000000"`
  - :val:`"cod/2000000@1234567"`
  - :val:`"nomad/L1234567890"`
  - :val:`"42"`""",
    )

    type: str = Field(
        ...,
        description="""The name of the type of an entry.
- **Type**: string.
- **Requirements/Conventions**:

  - **Support**: REQUIRED, MUST NOT be :val:`null`.
  - **Query**: MUST be a queryable property with support for all mandatory filter features.
  - **Response**: REQUIRED in the response.
  - MUST be an existing entry type.
  - The entry of type `<type>` and ID `<id>` MUST be returned in response to a request for :endpoint:`/<type>/<id>` under the versioned base URL.

- **Example**: :val:`"structures"`""",
    )

    attributes: EntryResourceAttributes = Field(
        ...,
        description="""a dictionary, containing key-value pairs representing the entry's properties, except for type and id.

Database-provider-specific properties need to include the database-provider-specific prefix
(see appendix `Database-Provider-Specific Namespace Prefixes`_).""",
    )

    relationships: Optional[EntryRelationships] = Field(
        None,
        description="""a dictionary containing references to other entries according to the description in section `Relationships`_
encoded as `JSON API Relationships <https://jsonapi.org/format/1.0/#document-resource-object-relationships>`__.
The OPTIONAL human-readable description of the relationship MAY be provided in the :field:`description` field inside the :field:`meta` dictionary.""",
    )


class EntryInfoProperty(BaseModel):

    description: str = Field(..., description="description of the entry property")

    unit: Optional[str] = Field(
        None, description="the physical unit of the entry property"
    )

    sortable: Optional[bool] = Field(
        None,
        description='defines whether the entry property can be used for sorting with the "sort" parameter. '
        "If the entry listing endpoint supports sorting, this key MUST be present for sortable properties with value `true`.",
    )


class EntryInfoResource(BaseModel):

    formats: List[str] = Field(..., description="list of available output formats.")

    description: str = Field(..., description="description of the entry")

    properties: Dict[str, EntryInfoProperty] = Field(
        ...,
        description="a dictionary describing queryable properties for this "
        "entry type, where each key is a property ID.",
    )

    output_fields_by_format: Dict[str, List[str]] = Field(
        ...,
        description="a dictionary of available output fields for this entry "
        "type, where the keys are the values of the `formats` list "
        "and the values are the keys of the `properties` dictionary.",
    )



import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='dmol3_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='dmol3.nomadmetainfo.json'))


class dmol3_section_hirshfeld_population(MSection):
    '''
    Hirshfeld Population Analysis Section
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='dmol3_section_hirshfeld_population'))

    dmol3_hirshfeld_population = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Hirshfeld Population Analysis
        ''',
        a_legacy=LegacyDefinition(name='dmol3_hirshfeld_population'))


class dmol3_section_mulliken_population(MSection):
    '''
    Mulliken Population Analysis Section
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='dmol3_section_mulliken_population'))

    dmol3_mulliken_population = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Mulliken Population Analysis
        ''',
        a_legacy=LegacyDefinition(name='dmol3_mulliken_population'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    dmol3_aux_density = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 aux density
        ''',
        a_legacy=LegacyDefinition(name='dmol3_aux_density'))

    dmol3_aux_partition = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 aux parition
        ''',
        a_legacy=LegacyDefinition(name='dmol3_aux_partition'))

    dmol3_basis_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 basis name
        ''',
        a_legacy=LegacyDefinition(name='dmol3_basis_name'))

    dmol3_calculation_type = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 calculation type
        ''',
        a_legacy=LegacyDefinition(name='dmol3_calculation_type'))

    dmol3_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 system charge
        ''',
        a_legacy=LegacyDefinition(name='dmol3_charge'))

    dmol3_electrostatic_moments = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Electrostatic_Moments
        ''',
        a_legacy=LegacyDefinition(name='dmol3_electrostatic_moments'))

    dmol3_functional_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 functional name
        ''',
        a_legacy=LegacyDefinition(name='dmol3_functional_name'))

    dmol3_hirshfeld_analysis = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Hirshfeld_Analysis
        ''',
        a_legacy=LegacyDefinition(name='dmol3_hirshfeld_analysis'))

    dmol3_integration_grid = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 integration grid
        ''',
        a_legacy=LegacyDefinition(name='dmol3_integration_grid'))

    dmol3_kpoints = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Kpoints
        ''',
        a_legacy=LegacyDefinition(name='dmol3_kpoints'))

    dmol3_mulliken_analysis = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Mulliken_Analysis
        ''',
        a_legacy=LegacyDefinition(name='dmol3_mulliken_analysis'))

    dmol3_nuclear_efg = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Nuclear_EFG
        ''',
        a_legacy=LegacyDefinition(name='dmol3_nuclear_efg'))

    dmol3_occupation_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Occupation name
        ''',
        a_legacy=LegacyDefinition(name='dmol3_occupation_name'))

    dmol3_occupation_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 Occupation width
        ''',
        a_legacy=LegacyDefinition(name='dmol3_occupation_width'))

    dmol3_opt_coordinate_system = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 OPT_Coordinate_System
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_coordinate_system'))

    dmol3_opt_displacement_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Displacement_Convergence
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_displacement_convergence'))

    dmol3_opt_energy_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Energy_Convergence
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_energy_convergence'))

    dmol3_opt_gdiis = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 OPT_Gdiis
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_gdiis'))

    dmol3_opt_gradient_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Gradient_Convergence
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_gradient_convergence'))

    dmol3_opt_hessian_project = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 OPT_Hessian_Project
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_hessian_project'))

    dmol3_opt_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 OPT_Iterations
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_iterations'))

    dmol3_opt_max_displacement = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Max_Displacement
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_max_displacement'))

    dmol3_opt_steep_tol = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Steep_Tol
        ''',
        a_legacy=LegacyDefinition(name='dmol3_opt_steep_tol'))

    dmol3_optical_absorption = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Optical_Absorption
        ''',
        a_legacy=LegacyDefinition(name='dmol3_optical_absorption'))

    dmol3_partial_dos = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Partial_Dos
        ''',
        a_legacy=LegacyDefinition(name='dmol3_partial_dos'))

    dmol3_pseudopotential_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 pseudopotential name
        ''',
        a_legacy=LegacyDefinition(name='dmol3_pseudopotential_name'))

    dmol3_rcut = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 atom R_cut
        ''',
        a_legacy=LegacyDefinition(name='dmol3_rcut'))

    dmol3_scf_charge_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_Charge_Mixing
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_charge_mixing'))

    dmol3_scf_density_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_Density_Convergence
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_density_convergence'))

    dmol3_scf_diis_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 SCF_DIIS name
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_diis_name'))

    dmol3_scf_diis_number = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_DIIS number
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_diis_number'))

    dmol3_scf_direct = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 SCF_Direct
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_direct'))

    dmol3_scf_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 SCF_Iterations
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_iterations'))

    dmol3_scf_number_bad_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 SCF_Number_Bad_Steps
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_number_bad_steps'))

    dmol3_scf_restart = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 SCF_Restart
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_restart'))

    dmol3_scf_spin_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_Spin_Mixing
        ''',
        a_legacy=LegacyDefinition(name='dmol3_scf_spin_mixing'))

    dmol3_spin_polarization = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 spin polarization
        ''',
        a_legacy=LegacyDefinition(name='dmol3_spin_polarization'))

    dmol3_spin = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 number of unpaired electrons
        ''',
        a_legacy=LegacyDefinition(name='dmol3_spin'))

    dmol3_symmetry = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 sysmmetry
        ''',
        a_legacy=LegacyDefinition(name='dmol3_symmetry'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    dmol3_binding_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        dmol3 binding energy at every SCF
        ''',
        a_legacy=LegacyDefinition(name='dmol3_binding_energy_scf_iteration'))

    dmol3_convergence_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 convergence at every SCF
        ''',
        a_legacy=LegacyDefinition(name='dmol3_convergence_scf_iteration'))

    dmol3_number_scf_iteration = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 iteration number at every SCF
        ''',
        a_legacy=LegacyDefinition(name='dmol3_number_scf_iteration'))

    dmol3_time_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 time at every SCF
        ''',
        a_legacy=LegacyDefinition(name='dmol3_time_scf_iteration'))


class section_eigenvalues(public.section_eigenvalues):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_eigenvalues'))

    dmol3_eigenvalue_eigenvalue = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Single eigenvalue
        ''',
        a_legacy=LegacyDefinition(name='dmol3_eigenvalue_eigenvalue'))

    dmol3_eigenvalue_occupation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Occupation of single eigenfunction
        ''',
        a_legacy=LegacyDefinition(name='dmol3_eigenvalue_occupation'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    dmol3_geometry_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        labels of atom
        ''',
        a_legacy=LegacyDefinition(name='dmol3_geometry_atom_labels'))

    dmol3_geometry_atom_positions_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of atomic position
        ''',
        a_legacy=LegacyDefinition(name='dmol3_geometry_atom_positions_x'))

    dmol3_geometry_atom_positions_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of atomic position
        ''',
        a_legacy=LegacyDefinition(name='dmol3_geometry_atom_positions_y'))

    dmol3_geometry_atom_positions_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of atomic position
        ''',
        a_legacy=LegacyDefinition(name='dmol3_geometry_atom_positions_z'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    dmol3_program_compilation_date = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 compilation date
        ''',
        a_legacy=LegacyDefinition(name='dmol3_program_compilation_date'))

    dmol3_program_compilation_time = Quantity(
        type=str,
        shape=[],
        description='''
        dmol compilation date
        ''',
        a_legacy=LegacyDefinition(name='dmol3_program_compilation_time'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    dmol3_section_hirshfeld_population = SubSection(
        sub_section=SectionProxy('dmol3_section_hirshfeld_population'),
        repeats=True,
        a_legacy=LegacyDefinition(name='dmol3_section_hirshfeld_population'))

    dmol3_section_mulliken_population = SubSection(
        sub_section=SectionProxy('dmol3_section_mulliken_population'),
        repeats=True,
        a_legacy=LegacyDefinition(name='dmol3_section_mulliken_population'))


m_package.__init_metainfo__()



import json
import sys
from pathlib import Path

try:
    from optimade import __api_version__
except ImportError:
    raise ImportError(
        "optimade needs to be installed prior to running 'optimade_version_update.py'"
    )

shields_json = Path(__file__).resolve().parent.joinpath("optimade-version.json")

with open(shields_json, "r") as fp:
    shield = json.load(fp)

shield_version = shield["message"]
current_version = f"v{__api_version__}"

if shield_version == current_version:
    # The shield has the newest implemented version
    print(
        f"""They are the same: {current_version}
Shield file:
{json.dumps(shield, indent=2)}"""
    )
    sys.exit(0)

print(
    f"""The shield version is outdated.
Shield version: {shield_version}
Current version: {current_version}
"""
)

shield["message"] = current_version
with open(shields_json, "w") as fp:
    json.dump(shield, fp, indent=2)
    fp.write("\n")

# Check file was saved correctly
with open(shields_json, "r") as fp:
    updated_shield = json.load(fp)

if updated_shield["message"] == current_version:
    print(f"Successfully updated the shield version to '{updated_shield['message']}'")
    sys.exit(0)
else:
    print(
        f"""Something went wrong !
Shield file:
{json.dumps(updated_shield, indent=2)}"""
    )
    sys.exit(1)



import os
import json
from nomad import infrastructure
from nomad.datamodel.material import Material, Similarity, DOSSimilarity
from nomad.archive import ArchiveReader, ArchiveWriter
from typing import List


def ingest(input_path: str, batch_size: int, verbose: bool):
    """Used to ingest the given DOS similarity values into MongoDB.

    Args:
        input_path: Path of the msgpack file to ingest.
        batch_size: Batch size for MongoDB bulk ingest.
        verbose: Enable verbose output.
    """
    # Initialize mongo connection
    infrastructure.setup_mongo()

    bulk = Material.m_def.a_mongo.mongo_cls()._get_collection().initialize_ordered_bulk_op()  # pylint: disable=not-callable

    with ArchiveReader(input_path) as reader:
        i = 0
        for material_id in reader:
            material_object = reader[material_id]
            material_dict = material_object.to_dict()

            # Each entry is cycled through the material metainfo definition and
            # the mongo annotation to validate it and only push the annotated
            # data.
            mongo_instance = Material.m_def.a_mongo.mongo_cls(**material_dict, _created=False)  # pylint: disable=not-callable
            material_dict = mongo_instance.to_mongo().to_dict()

            # Add an upsert to the bulk operations
            bulk.find({'_id': material_dict["_id"]}).upsert().update_one({'$set': material_dict})
            i += 1
            if i % batch_size == 0:
                bulk.execute()
                bulk = Material.m_def.a_mongo.mongo_cls()._get_collection().initialize_ordered_bulk_op()  # pylint: disable=not-callable
                if verbose:
                    print("{} inserted".format(i))

    # Final push for the remainder
    bulk.execute()
    if verbose:
        print("{} inserted".format(i))


def update(input_dir: str, output_path: str, verbose: bool):
    """Used to create a compact msgpack file that follows the metainfo schema
    for materials and contains the DOS similarity values.

    Args:
        input_dir: Path of the directory containing the raw similarity filesa.
        output_path: Path of the output msgpack file.
        verbose: Enable verbose output.
    """
    # Find all valid data files in the given directory
    similarity_files: List[str] = []
    for filename in os.listdir(input_dir):
        if filename.endswith(".dat"):
            similarity_files.append(os.path.join(input_dir, filename))
    n_files: int = len(similarity_files)
    if n_files == 0:
        raise ValueError("Could not find similarity files in directory: {}".format(input_dir))
    if verbose:
        print("{} files found".format(n_files))

    # Gather the number of entries to prepare the msgpack file.
    n_entries = 0
    for filepath in similarity_files:
        num_lines = sum(1 for line in open(filepath, "r"))
        n_entries += num_lines
    if verbose:
        print("{} entries found".format(n_entries))

    # Read the data file containing similarities
    # i = 0
    if verbose:
        print("Writing msgpack file...")
    with ArchiveWriter(output_path, n_entries, entry_toc_depth=1) as writer:
        for filepath in similarity_files:
            with open(filepath, "r") as f:
                for line in f:
                    ientry = json.loads(line)

                    for key, value in ientry.items():
                        _, _, imaterial = key.split(":")

                        # Create data according to a metainfo model
                        material = Material()
                        material.material_id = imaterial
                        similarity = material.m_create(Similarity)
                        dos_similarity = similarity.m_create(DOSSimilarity)
                        ids = []
                        values = []

                        for jkey, similarity_value in value.items():
                            _, _, jmaterial = jkey.split(":")
                            ids.append(jmaterial)
                            values.append(similarity_value)

                        dos_similarity.material_ids = ids
                        dos_similarity.values = values

                        # Save as msgpack
                        writer.add(imaterial, material.m_to_dict())

    if verbose:
        print("Finished")



# Copyright 2016-2018 Arvid Conrad Ihrig, Aliaksei Mazheika
#                     Fritz-Haber-Institut der Max-Planck-Gesellschaft
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import numpy as np
from nomadcore.simple_parser import SimpleMatcher as SM
from turbomoleparser.TurbomoleCommon import RE_FLOAT

logger = logging.getLogger("nomad.turbomoleParser")


class MethodParser(object):

    __functional_map = {
        "S-VWN":  ["LDA_X", "LDA_C_VWN_3"],
        "PWLDA":  ["LDA_X", "LDA_C_PW"],
        "B-VWN":  ["GGA_X_B88", "LDA_C_VWN"],
        "B-LYP":  ["GGA_X_B88", "GGA_C_LYP"],
        "B-P":    ["GGA_X_B88", "GGA_C_P86"],
        "B-P86":  ["GGA_X_B88", "GGA_C_P86"],
        "PBE":    ["GGA_X_PBE", "GGA_C_PBE"],
        "TPSS":   ["MGGA_X_TPSS", "MGGA_C_TPSS"],
        "M06":    ["MGGA_X_M06", "MGGA_C_M06"],
        "BH-LYP": ["HYB_GGA_XC_BHANDHLYP"],
        "B3-LYP": ["HYB_GGA_XC_B3LYP"],
        "PBE0":   ["HYB_GGA_XC_PBEH"],
        "TPSSh":  ["HYB_MGGA_XC_TPSSH"],
        "M06-2X": ["MGGA_X_M06_2X", "MGGA_C_M06_2X"],
        "B2-PLYP":["HYB_GGA_XC_B2PLYP"]
    }

    # TODO: verify mapping of approximations to general Electronic Structure Method class
    __wavefunction_models_map = {
        "MP2": "MP2",
        "CCS": "CCS",
        "CIS": "CIS",
        "CIS(D)": "CISD",
        "CIS(Dinf)": "CISD",
        "ADC(2)": "MP2",  # TODO: check paper to verify this mapping
        "CC2": "CCSD",
        "CCSD": "CCSD",
        "CCSD(T)": "CCSD(T)"
    }

    def __init__(self, context, key="method"):
        context[key] = self
        self.__context = context
        self.__backend = None
        self.spin_channels = 1
        self.k_points = 1
        self.__method = None
        self.__functional = None
        self.__energy_kinetic = None
        self.__energy_potential = None
        self.__indices_to_write = {"method": -1, "config": -1}

    def purge_data(self):
        self.spin_channels = 1
        self.k_points = 1
        self.__method = None
        self.__functional = None
        self.__energy_kinetic = None
        self.__energy_potential = None

    def set_backend(self, backend):
        self.__backend = backend

    def get_energy_kinetic(self):
        return self.__energy_kinetic

    def get_energy_potential(self):
        return self.__energy_potential

    def get_main_method(self):
        return self.__method

    # matcher generation methods

    def add_default_functional(self, backend, gIndex, section):
        if not self.__method:
            index_method = self.__context.index_method()
            self.__method = "DFT"
            self.__backend.addValue("electronic_structure_method", "DFT", index_method)
            self.__backend.addValue("calculation_method_kind", "absolute", index_method)
            index_xc = self.__backend.openSection("section_XC_functionals")
            self.__backend.addValue('XC_functional_name', "HF_X", index_xc)
            self.__backend.closeSection("section_XC_functionals", index_xc)

    def build_uhf_matcher(self):

        def set_spin_polarized(backend, groups):
            self.__spin_channels = 2

        return SM(r"\s*UHF mode switched on !",
                  name="UHF switch",
                  startReAction=set_spin_polarized
                  )

    # TODO: add support for remaining XC-functionals in Turbomole + custom mixing combinations
    def build_dft_functional_matcher(self):
        def set_exchange_part(backend, groups):
            backend.addValue("x_turbomole_functional_type_exchange", groups[0],
                             self.__context.index_method())

        def set_correlation_part(backend, groups):
            backend.addValue("x_turbomole_functional_type_correlation", groups[0],
                             self.__context.index_method())

        exchange = SM(r"\s*exchange:\s*(.+)",
                      name="exchange part",
                      startReAction=set_exchange_part
                      )
        correlation = SM(r"\s*correlation:\s*(.+)",
                         name="exchange part",
                         startReAction=set_correlation_part
                         )

        def set_functional(backend, groups):
            index_method = self.__context.index_method()
            if groups[0] in self.__functional_map:
                self.__functional = self.__functional_map[groups[0]]
            else:
                self.__functional = ["UNKNOWN"]
                logger.warning("XC-functional '%s' not known!" % groups[0])
            self.__method = "DFT"
            backend.addValue("electronic_structure_method", "DFT", index_method)
            backend.addValue("calculation_method_kind", "absolute", index_method)
            for component in self.__functional:
                index_xc = backend.openSection("section_XC_functionals")
                backend.addValue('XC_functional_name', component, index_xc)
                backend.closeSection("section_XC_functionals", index_xc)

        return SM(r"\s*density functional\s*$",
                  name="DFT functional",
                  subMatchers=[
                      SM(r"\s*-{5,}\s*$",
                         name="<format>",
                         coverageIgnore=True
                         ),
                      SM(r"\s*([A-z0-9-]+)\s+functional",
                         name="XC Functional",
                         startReAction=set_functional
                         ),
                      SM(r"\s*([A-z0-9-]+)\s+meta-GGA functional",
                         name="XC Functional",
                         startReAction=set_functional
                         ),
                      SM(r".*functional\s*:\s*([A-z0-9-]+)",
                         name="XC Functional",
                         startReAction=set_functional
                         ),
                      exchange,
                      correlation
                  ]
                  )

    def build_wave_function_model_matcher(self):

        def determine_spin(backend, groups):
            if groups[0] == "restricted closed":
                pass
            elif groups[0] == "restricted open":
                self.__context["method"].spin_channels = 2
            elif groups[0] == "unrestricted open":
                self.__context["method"].spin_channels = 2
            else:
                logger.error("found unknown spin configuration in : %s" % groups[0])

        def extract_wf_method(backend, groups):
            method = self.__wavefunction_models_map.get(groups[0], None)
            self.__method = groups[0]
            if method:
                backend.addValue("electronic_structure_method", method,
                                 self.__context.index_method())
            else:
                logger.error("unknown wave-function model encountered: %s - %s" % groups)

        method_matcher = SM(r"\s*([^\s].*[^\s])\s*-\s*([^\s].*[^\s])\s*$",
                            name="WF model",
                            required=True,
                            startReAction=extract_wf_method
                            )

        # TODO: extract further RICC2 parameters?
        return SM(r"\s*([^ ].+)\s+shell\s+calculation\s+for\s+the\s+wavefunction\s+models:\s*$",
                  startReAction=determine_spin,
                  name="spin treatment",
                  required=True,
                  subMatchers=[
                      method_matcher
                  ]
                  )

    def build_dftd3_vdw_matcher(self):

        def store_energy(backend, groups):
            backend.addRealValue("energy_van_der_Waals", float(groups[0]),
                                 self.__context.index_configuration(), unit="hartree")

        def store_version(backend, groups):
            backend.addValue("x_turbomole_dft_d3_version", groups[0], self.__context.index_method())
            backend.addValue("van_der_Waals_method", "DFT-D3", self.__context.index_method())

        energy_matcher = SM(r"\s*Edisp\s+/kcal,\s*au\s*:\s*"+RE_FLOAT+"\s+("+RE_FLOAT+")\s*$",
                            name="vdW energy",
                            startReAction=store_energy
                            )

        return SM(r"\s*\|\s*DFTD3\s+(V[0-9]+.[0-9]+\s+Rev\s+[0-9]+)\s*\|\s*$",
                  name="DFT-D3 version",
                  startReAction=store_version,
                  subMatchers=[
                      energy_matcher
                  ],
                  )

    def build_total_energy_matcher(self):
        def set_current_energy(backend, groups):
            backend.addRealValue("energy_current", float(groups[0]),
                                 self.__context.index_configuration(), unit="hartree")
            backend.addRealValue("energy_total", float(groups[0]),
                                 self.__context.index_configuration(), unit="hartree")

        def store_kinetic_energy(backend, groups):
            self.__energy_kinetic = groups[0]
            backend.addRealValue("electronic_kinetic_energy", float(groups[0]),
                                 self.__context.index_configuration(), unit="hartree")

        def store_kinetic_potential(backend, groups):
            self.__energy_potential = groups[0]
            backend.addRealValue("x_turbomole_potential_energy_final", float(groups[0]),
                                 self.__context.index_configuration(), unit="hartree")

        def store_wavefunc_norm(backend, groups):
            backend.addRealValue("x_turbomole_wave_func_norm", float(groups[0]),
                                 self.__context.index_configuration())

        def store_virial_theorem(backend, groups):
            backend.addRealValue("x_turbomole_virial_theorem", float(groups[0]),
                                 self.__context.index_configuration())

        def store_scf_convergence(backend, groups):
            backend.addValue("number_of_scf_iterations", int(groups[0]),
                             self.__context.index_configuration())

        energy_total = SM(r"\s*\|\s*total energy\s*=\s*("+RE_FLOAT+")\s*\|",
                          name="total energy",
                          required=True,
                          startReAction=set_current_energy
                          )
        energy_kinetic = SM(r"\s*:\s*kinetic energy\s*=\s*("+RE_FLOAT+")\s*:\s*$",
                            name="kinetic energy",
                            required=True,
                            startReAction=store_kinetic_energy
                            )
        energy_potential = SM(r"\s*:\s*potential energy\s*=\s*("+RE_FLOAT+")\s*:\s*$",
                              name="potential energy",
                              required=True,
                              startReAction=store_kinetic_potential
                              )
        virial_theorem = SM(r"\s*\:\s*virial theorem\s*\=\s*("+RE_FLOAT+")\s*:\s*$",
                            name="virial theorem",
                            startReAction=store_virial_theorem,
                            required=True
                            )
        wavefunction_norm = SM(r"\s*\:\s*wavefunction norm\s*\=\s*("+RE_FLOAT+")\s*:\s*$",
                               name="wavefunction norm",
                               startReAction=store_wavefunc_norm,
                               required=True
                               )

        return SM(r"\s*convergence criteria satisfied after\s+([0-9]+)\s+iterations",
                  name="SCF end",
                  startReAction=store_scf_convergence,
                  required=True,
                  subMatchers=[
                      energy_total,
                      energy_kinetic,
                      energy_potential,
                      virial_theorem,
                      wavefunction_norm
                  ]
                  )

    def build_correlation_energy_matcher(self):
        def correlation_correction(backend, groups):
            backend.addRealValue("energy_current", float(groups[0]),
                                 self.__context.index_configuration(), unit="hartree")

        def correlated_total_energy(backend, groups):
            if groups[0] != self.__method:
                backend.addValue("electronic_structure_method", groups[0],
                                 self.__context.index_method())
            backend.addRealValue("energy_total", float(groups[1]),
                                 self.__context.index_configuration(), unit="hartree")

        correlation = SM(r"\s*\*\s*(?:total\s+)?correlation\s+energy\s*:\s*("+RE_FLOAT+r")\s*\*\s*",
                         name="correlation energy",
                         startReAction=correlation_correction
                         )
        mp2_correlation = SM(r"\s*\*\s*MP2\s+correlation\s+energy\s+\(doubles\)\s*:\s*("
                             + RE_FLOAT + r")\s*\*\s*",
                             name="correlation energy",
                             startReAction=correlation_correction
                             )
        total_energy = SM(r"\s*\*\s*Final\s+([^\s].+[^\s])\s+energy\s*:\s*("+RE_FLOAT+r")\s*\*\s*",
                          name="total energy",
                          startReAction=correlated_total_energy
                          )

        return SM(r"\s*\*\s*(RHF|UHF|ROHF)\s+energy\s*:\s*("+RE_FLOAT+r")\s*\*\s*$",
                  name="HF energy",
                  subMatchers=[
                      mp2_correlation,
                      correlation,
                      total_energy
                  ]
                  )



# pylint: disable=import-error

import pytest

from .utils import get_min_ver

min_ver = get_min_ver("aiida-core")
aiida = pytest.importorskip(
    "aiida",
    minversion=min_ver,
    reason=f"aiida-core must be installed with minimum version {min_ver} for these tests to"
    " be able to run",
)

from aiida import load_profile

load_profile()

from aiida.orm import StructureData

from optimade.adapters import Structure
from optimade.adapters.structures.aiida import get_aiida_structure_data


def test_successful_conversion(RAW_STRUCTURES):
    """Make sure its possible to convert"""
    for structure in RAW_STRUCTURES:
        assert isinstance(get_aiida_structure_data(Structure(structure)), StructureData)


def test_null_positions(null_position_structure):
    """Make sure null positions are handled"""
    assert isinstance(get_aiida_structure_data(null_position_structure), StructureData)


def test_null_lattice_vectors(null_lattice_vector_structure):
    """Make sure null lattice vectors are handled"""
    assert isinstance(
        get_aiida_structure_data(null_lattice_vector_structure), StructureData
    )


def test_special_species(SPECIAL_SPECIES_STRUCTURES):
    """Make sure vacancies and non-chemical symbols ("X") are handled"""
    for special_structure in SPECIAL_SPECIES_STRUCTURES:
        structure = Structure(special_structure)

        aiida_structure = get_aiida_structure_data(structure)

        assert isinstance(aiida_structure, StructureData)

        if "vacancy" in structure.attributes.species[0].chemical_symbols:
            assert aiida_structure.has_vacancies
            assert not aiida_structure.is_alloy
        elif len(structure.attributes.species[0].chemical_symbols) > 1:
            assert not aiida_structure.has_vacancies
            assert aiida_structure.is_alloy
        else:
            assert not aiida_structure.has_vacancies
            assert not aiida_structure.is_alloy
